import math
import logging
from enum import Enum
from typing import Optional, List, Tuple, Any, Union, Dict, Callable
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO

import requests
import numpy as np
from google.api_core import retry
from PIL import Image
from pyproj import Transformer
from pygeotile.point import Point as PygeoPoint

from labelbox.data.annotation_types import Rectangle, Point, Line, Polygon
from .base_data import BaseData
from .raster import RasterData
from pydantic import BaseModel, field_validator, model_validator, ConfigDict

VALID_LAT_RANGE = range(-90, 90)
VALID_LNG_RANGE = range(-180, 180)
DEFAULT_TMS_TILE_SIZE = 256
TILE_DOWNLOAD_CONCURRENCY = 4

logger = logging.getLogger(__name__)

VectorTool = Union[Point, Line, Rectangle, Polygon]


class EPSG(Enum):
    """Provides the EPSG for tiled image assets that are currently supported.

    SIMPLEPIXEL is Simple that can be used to obtain the pixel space coordinates

    >>> epsg = EPSG()
    """

    SIMPLEPIXEL = 1
    EPSG4326 = 4326
    EPSG3857 = 3857


class TiledBounds(BaseModel):
    """Bounds for a tiled image asset related to the relevant epsg.

    Bounds should be Point objects.

    >>> bounds = TiledBounds(epsg=EPSG.EPSG4326,
            bounds=[
                Point(x=-99.21052827588443, y=19.405662413477728),
                Point(x=-99.20534818927473, y=19.400498983095076)
            ])
    """

    epsg: EPSG
    bounds: List[Point]

    @field_validator("bounds")
    def validate_bounds_not_equal(cls, bounds):
        first_bound = bounds[0]
        second_bound = bounds[1]

        if first_bound.x == second_bound.x or first_bound.y == second_bound.y:
            raise ValueError(
                f"Bounds on either axes cannot be equal, currently {bounds}"
            )
        return bounds

    # validate bounds are within lat,lng range if they are EPSG4326
    @model_validator(mode="after")
    def validate_bounds_lat_lng(self):
        epsg = self.epsg
        bounds = self.bounds

        if epsg == EPSG.EPSG4326:
            for bound in bounds:
                lat, lng = bound.y, bound.x
                if (
                    int(lng) not in VALID_LNG_RANGE
                    or int(lat) not in VALID_LAT_RANGE
                ):
                    raise ValueError(
                        f"Invalid lat/lng bounds. Found {bounds}. "
                        f"lat must be in {VALID_LAT_RANGE}. "
                        f"lng must be in {VALID_LNG_RANGE}."
                    )
        return self


class TileLayer(BaseModel):
    """Url that contains the tile layer. Must be in the format:

    https://c.tile.openstreetmap.org/{z}/{x}/{y}.png

    >>> layer = TileLayer(
        url="https://c.tile.openstreetmap.org/{z}/{x}/{y}.png",
        name="slippy map tile"
        )
    """

    url: str
    name: Optional[str] = "default"

    def asdict(self) -> Dict[str, str]:
        return {"tileLayerUrl": self.url, "name": self.name}

    @field_validator("url")
    def validate_url(cls, url):
        xyz_format = "/{z}/{x}/{y}"
        if xyz_format not in url:
            raise ValueError(f"{url} needs to contain {xyz_format}")
        return url


class TiledImageData(BaseData):
    """Represents tiled imagery

    If specified version is 2, converts bounds from [lng,lat] to [lat,lng]

    Requires the following args:
        tile_layer: TileLayer
        tile_bounds: TiledBounds
        zoom_levels: List[int]
    Optional args:
        max_native_zoom: int = None
        tile_size: Optional[int]
        version: int = 2
        alternative_layers: List[TileLayer]

    >>> tiled_image_data = TiledImageData(tile_layer=TileLayer,
                                  tile_bounds=TiledBounds,
                                  zoom_levels=[1, 12])
    """

    tile_layer: TileLayer
    tile_bounds: TiledBounds
    alternative_layers: List[TileLayer] = []
    zoom_levels: Tuple[int, int]
    max_native_zoom: Optional[int] = None
    tile_size: Optional[int] = DEFAULT_TMS_TILE_SIZE
    version: Optional[int] = 2
    multithread: bool = True

    def __post_init__(self) -> None:
        if self.max_native_zoom is None:
            self.max_native_zoom = self.zoom_levels[0]

    def asdict(self) -> Dict[str, str]:
        return {
            "tileLayerUrl": self.tile_layer.url,
            "bounds": [
                [self.tile_bounds.bounds[0].x, self.tile_bounds.bounds[0].y],
                [self.tile_bounds.bounds[1].x, self.tile_bounds.bounds[1].y],
            ],
            "minZoom": self.zoom_levels[0],
            "maxZoom": self.zoom_levels[1],
            "maxNativeZoom": self.max_native_zoom,
            "epsg": self.tile_bounds.epsg.name,
            "tileSize": self.tile_size,
            "alternativeLayers": [
                layer.asdict() for layer in self.alternative_layers
            ],
            "version": self.version,
        }

    def raster_data(
        self, zoom: int = 0, max_tiles: int = 32, multithread=True
    ) -> RasterData:
        """Converts the tiled image asset into a RasterData object containing an
        np.ndarray.

        Uses the minimum zoom provided to render the image.
        """
        if self.tile_bounds.epsg == EPSG.SIMPLEPIXEL:
            xstart, ystart, xend, yend = self._get_simple_image_params(zoom)
        elif self.tile_bounds.epsg == EPSG.EPSG4326:
            xstart, ystart, xend, yend = self._get_3857_image_params(
                zoom, self.tile_bounds
            )
        elif self.tile_bounds.epsg == EPSG.EPSG3857:
            # transform to 4326
            transformer = EPSGTransformer.create_geo_to_geo_transformer(
                EPSG.EPSG3857, EPSG.EPSG4326
            )
            transforming_bounds = [
                transformer(self.tile_bounds.bounds[0]),
                transformer(self.tile_bounds.bounds[1]),
            ]
            xstart, ystart, xend, yend = self._get_3857_image_params(
                zoom, transforming_bounds
            )
        else:
            raise ValueError(f"Unsupported epsg found: {self.tile_bounds.epsg}")

        self._validate_num_tiles(xstart, ystart, xend, yend, max_tiles)

        rounded_tiles, pixel_offsets = list(
            zip(
                *[
                    self._tile_to_pixel(pt)
                    for pt in [xstart, ystart, xend, yend]
                ]
            )
        )

        image = self._fetch_image_for_bounds(*rounded_tiles, zoom, multithread)
        arr = self._crop_to_bounds(image, *pixel_offsets)
        return RasterData(arr=arr)

    @property
    def value(self) -> np.ndarray:
        """Returns the value of a generated RasterData object."""
        return self.raster_data(
            self.zoom_levels[0], multithread=self.multithread
        ).value

    def _get_simple_image_params(
        self, zoom
    ) -> Tuple[float, float, float, float]:
        """Computes the x and y tile bounds for fetching an image that
        captures the entire labeling region (TiledData.bounds) given a specific zoom

        Simple has different order of x / y than lat / lng because of how leaflet behaves
        leaflet reports all points as pixel locations at a zoom of 0
        """
        xend, xstart, yend, ystart = (
            self.tile_bounds.bounds[1].x,
            self.tile_bounds.bounds[0].x,
            self.tile_bounds.bounds[1].y,
            self.tile_bounds.bounds[0].y,
        )
        return (
            *[
                x * (2 ** (zoom)) / self.tile_size
                for x in [xstart, ystart, xend, yend]
            ],
        )

    def _get_3857_image_params(
        self, zoom: int, bounds: TiledBounds
    ) -> Tuple[float, float, float, float]:
        """Computes the x and y tile bounds for fetching an image that
        captures the entire labeling region (TiledData.bounds) given a specific zoom
        """
        lat_start, lat_end = bounds.bounds[1].y, bounds.bounds[0].y
        lng_start, lng_end = bounds.bounds[1].x, bounds.bounds[0].x

        # Convert to zoom 0 tile coordinates
        xstart, ystart = self._latlng_to_tile(lat_start, lng_start)
        xend, yend = self._latlng_to_tile(lat_end, lng_end)

        # Make sure that the tiles are increasing in order
        xstart, xend = min(xstart, xend), max(xstart, xend)
        ystart, yend = min(ystart, yend), max(ystart, yend)
        return (*[pt * 2.0**zoom for pt in [xstart, ystart, xend, yend]],)

    def _latlng_to_tile(
        self, lat: float, lng: float, zoom=0
    ) -> Tuple[float, float]:
        """Converts lat/lng to 3857 tile coordinates
        Formula found here:
        https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#lon.2Flat_to_tile_numbers_2
        """
        scale = 2**zoom
        lat_rad = math.radians(lat)
        x = (lng + 180.0) / 360.0 * scale
        y = (1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * scale
        return x, y

    def _tile_to_pixel(self, tile: float) -> Tuple[int, int]:
        """Rounds a tile coordinate and reports the remainder in pixels"""
        rounded_tile = int(tile)
        remainder = tile - rounded_tile
        pixel_offset = int(self.tile_size * remainder)
        return rounded_tile, pixel_offset

    def _fetch_image_for_bounds(
        self,
        x_tile_start: int,
        y_tile_start: int,
        x_tile_end: int,
        y_tile_end: int,
        zoom: int,
        multithread=True,
    ) -> np.ndarray:
        """Fetches the tiles and combines them into a single image.

        If a tile cannot be fetched, a padding of expected tile size is instead added.
        """

        if multithread:
            tiles = {}
            with ThreadPoolExecutor(
                max_workers=TILE_DOWNLOAD_CONCURRENCY
            ) as exc:
                for x in range(x_tile_start, x_tile_end + 1):
                    for y in range(y_tile_start, y_tile_end + 1):
                        tiles[(x, y)] = exc.submit(self._fetch_tile, x, y, zoom)

        rows = []
        for y in range(y_tile_start, y_tile_end + 1):
            row = []
            for x in range(x_tile_start, x_tile_end + 1):
                try:
                    if multithread:
                        row.append(tiles[(x, y)].result())
                    else:
                        row.append(self._fetch_tile(x, y, zoom))
                except:
                    row.append(
                        np.zeros(
                            shape=(self.tile_size, self.tile_size, 3),
                            dtype=np.uint8,
                        )
                    )
            rows.append(np.hstack(row))

        return np.vstack(rows)

    @retry.Retry(initial=1, maximum=16, multiplier=2)
    def _fetch_tile(self, x: int, y: int, z: int) -> np.ndarray:
        """
        Fetches the image and returns an np array.
        """
        data = requests.get(self.tile_layer.url.format(x=x, y=y, z=z))
        data.raise_for_status()
        decoded = np.array(Image.open(BytesIO(data.content)))[..., :3]
        if decoded.shape[:2] != (self.tile_size, self.tile_size):
            logger.warning(f"Unexpected tile size {decoded.shape}.")
        return decoded

    def _crop_to_bounds(
        self,
        image: np.ndarray,
        x_px_start: int,
        y_px_start: int,
        x_px_end: int,
        y_px_end: int,
    ) -> np.ndarray:
        """This function slices off the excess pixels that are outside of the bounds.
        This occurs because only full tiles can be downloaded at a time.
        """

        def invert_point(pt):
            # Must have at least 1 pixel for stability.
            pt = max(pt, 1)
            # All pixel points are relative to a single tile
            # So subtracting the tile size inverts the axis
            pt = pt - self.tile_size
            return pt if pt != 0 else None

        x_px_end, y_px_end = invert_point(x_px_end), invert_point(y_px_end)
        return image[y_px_start:y_px_end, x_px_start:x_px_end, :]

    def _validate_num_tiles(
        self,
        xstart: float,
        ystart: float,
        xend: float,
        yend: float,
        max_tiles: int,
    ):
        """Calculates the number of expected tiles we would fetch.

        If this is greater than the number of max tiles, raise an error.
        """
        total_n_tiles = (yend - ystart + 1) * (xend - xstart + 1)
        if total_n_tiles > max_tiles:
            raise ValueError(
                f"Requested zoom results in {total_n_tiles} tiles."
                f"Max allowed tiles are {max_tiles}"
                f"Increase max tiles or reduce zoom level."
            )

    @field_validator("zoom_levels")
    def validate_zoom_levels(cls, zoom_levels):
        if zoom_levels[0] > zoom_levels[1]:
            raise ValueError(
                f"Order of zoom levels should be min, max. Received {zoom_levels}"
            )
        return zoom_levels


class EPSGTransformer(BaseModel):
    """Transformer class between different EPSG's. Useful when wanting to project
    in different formats.
    """

    transformer: Any
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @staticmethod
    def _is_simple(epsg: EPSG) -> bool:
        return epsg == EPSG.SIMPLEPIXEL

    @staticmethod
    def _get_ranges(bounds: np.ndarray) -> Tuple[int, int]:
        """helper function to get the range between bounds.

        returns a tuple (x_range, y_range)"""
        x_range = np.max(bounds[:, 0]) - np.min(bounds[:, 0])
        y_range = np.max(bounds[:, 1]) - np.min(bounds[:, 1])
        return (x_range, y_range)

    @staticmethod
    def _min_max_x_y(bounds: np.ndarray) -> Tuple[int, int, int, int]:
        """returns the min x, max x, min y, max y of a numpy array"""
        return (
            np.min(bounds[:, 0]),
            np.max(bounds[:, 0]),
            np.min(bounds[:, 1]),
            np.max(bounds[:, 1]),
        )

    @classmethod
    def geo_and_pixel(
        cls,
        src_epsg,
        pixel_bounds: TiledBounds,
        geo_bounds: TiledBounds,
        zoom=0,
    ) -> Callable:
        """method to change from one projection to simple projection"""

        pixel_bounds = pixel_bounds.bounds
        geo_bounds_epsg = geo_bounds.epsg
        geo_bounds = geo_bounds.bounds

        local_bounds = np.array(
            [(point.x, point.y) for point in pixel_bounds], dtype=int
        )
        # convert geo bounds to pixel bounds. assumes geo bounds are in wgs84/EPS4326 per leaflet
        global_bounds = np.array(
            [
                PygeoPoint.from_latitude_longitude(
                    latitude=point.y, longitude=point.x
                ).pixels(zoom)
                for point in geo_bounds
            ]
        )

        # get the range of pixels for both sets of bounds to use as a multiplification factor
        local_x_range, local_y_range = cls._get_ranges(bounds=local_bounds)
        global_x_range, global_y_range = cls._get_ranges(bounds=global_bounds)

        if src_epsg == EPSG.SIMPLEPIXEL:

            def transform(x: int, y: int) -> Callable[[int, int], Transformer]:
                scaled_xy = (
                    x * (global_x_range) / (local_x_range),
                    y * (global_y_range) / (local_y_range),
                )

                minx, _, miny, _ = cls._min_max_x_y(bounds=global_bounds)
                x, y = map(lambda i, j: i + j, scaled_xy, (minx, miny))

                point = PygeoPoint.from_pixel(
                    pixel_x=x, pixel_y=y, zoom=zoom
                ).latitude_longitude
                # convert to the desired epsg
                return Transformer.from_crs(
                    EPSG.EPSG4326.value, geo_bounds_epsg.value, always_xy=True
                ).transform(point[1], point[0])

            return transform

        # handles 4326 from lat,lng
        elif src_epsg == EPSG.EPSG4326:

            def transform(x: int, y: int) -> Callable[[int, int], Transformer]:
                point_in_px = PygeoPoint.from_latitude_longitude(
                    latitude=y, longitude=x
                ).pixels(zoom)

                minx, _, miny, _ = cls._min_max_x_y(global_bounds)
                x, y = map(lambda i, j: i - j, point_in_px, (minx, miny))

                return (
                    x * (local_x_range) / (global_x_range),
                    y * (local_y_range) / (global_y_range),
                )

            return transform

        # handles 3857 from meters
        elif src_epsg == EPSG.EPSG3857:

            def transform(x: int, y: int) -> Callable[[int, int], Transformer]:
                point_in_px = PygeoPoint.from_meters(
                    meter_y=y, meter_x=x
                ).pixels(zoom)

                minx, _, miny, _ = cls._min_max_x_y(global_bounds)
                x, y = map(lambda i, j: i - j, point_in_px, (minx, miny))

                return (
                    x * (local_x_range) / (global_x_range),
                    y * (local_y_range) / (global_y_range),
                )

            return transform

    @classmethod
    def create_geo_to_geo_transformer(
        cls, src_epsg: EPSG, tgt_epsg: EPSG
    ) -> Callable[[int, int], Transformer]:
        """method to change from one projection to another projection.

        supports EPSG transformations not Simple.
        """
        if cls._is_simple(epsg=src_epsg) or cls._is_simple(epsg=tgt_epsg):
            raise Exception(
                f"Cannot be used for Simple transformations. Found {src_epsg} and {tgt_epsg}"
            )

        return EPSGTransformer(
            transformer=Transformer.from_crs(
                src_epsg.value, tgt_epsg.value, always_xy=True
            ).transform
        )

    @classmethod
    def create_geo_to_pixel_transformer(
        cls,
        src_epsg,
        pixel_bounds: TiledBounds,
        geo_bounds: TiledBounds,
        zoom=0,
    ) -> Callable[[int, int], Transformer]:
        """method to change from a geo projection to Simple"""

        transform_function = cls.geo_and_pixel(
            src_epsg=src_epsg,
            pixel_bounds=pixel_bounds,
            geo_bounds=geo_bounds,
            zoom=zoom,
        )
        return EPSGTransformer(transformer=transform_function)

    @classmethod
    def create_pixel_to_geo_transformer(
        cls,
        src_epsg,
        pixel_bounds: TiledBounds,
        geo_bounds: TiledBounds,
        zoom=0,
    ) -> Callable[[int, int], Transformer]:
        """method to change from a geo projection to Simple"""
        transform_function = cls.geo_and_pixel(
            src_epsg=src_epsg,
            pixel_bounds=pixel_bounds,
            geo_bounds=geo_bounds,
            zoom=zoom,
        )
        return EPSGTransformer(transformer=transform_function)

    def _get_point_obj(self, point) -> Point:
        point = self.transformer(point.x, point.y)
        return Point(x=point[0], y=point[1])

    def __call__(
        self, shape: Union[Point, Line, Rectangle, Polygon]
    ) -> Union[VectorTool, List[VectorTool]]:
        if isinstance(shape, list):
            return [self(geom) for geom in shape]
        if isinstance(shape, Point):
            return self._get_point_obj(shape)
        if isinstance(shape, Line):
            return Line(points=[self._get_point_obj(p) for p in shape.points])
        if isinstance(shape, Polygon):
            return Polygon(
                points=[self._get_point_obj(p) for p in shape.points]
            )
        if isinstance(shape, Rectangle):
            return Rectangle(
                start=self._get_point_obj(shape.start),
                end=self._get_point_obj(shape.end),
            )
        else:
            raise ValueError(f"Unsupported type found: {type(shape)}")



from typing import Any, Dict, List, Union, Optional

from labelbox.data.annotation_types import GenericDataRowData
from labelbox.data.mixins import (
    ConfidenceMixin,
    CustomMetric,
    CustomMetricsMixin,
)
from labelbox.data.serialization.ndjson.base import DataRow, NDAnnotation

from ....annotated_types import Cuid

from ...annotation_types.annotation import ClassificationAnnotation
from ...annotation_types.video import VideoClassificationAnnotation
from ...annotation_types.llm_prompt_response.prompt import (
    PromptClassificationAnnotation,
    PromptText,
)
from ...annotation_types.classification.classification import (
    ClassificationAnswer,
    Text,
    Checklist,
    Radio,
)
from pydantic import (
    model_validator,
    Field,
    BaseModel,
    ConfigDict,
    model_serializer,
)
from pydantic.alias_generators import to_camel


class NDAnswer(ConfidenceMixin, CustomMetricsMixin):
    name: Optional[str] = None
    schema_id: Optional[Cuid] = None
    classifications: Optional[List["NDSubclassificationType"]] = None
    model_config = ConfigDict(populate_by_name=True, alias_generator=to_camel)

    @model_validator(mode="after")
    def must_set_one(self):
        if (not hasattr(self, "schema_id") or self.schema_id is None) and (
            not hasattr(self, "name") or self.name is None
        ):
            raise ValueError("Schema id or name are not set. Set either one.")
        return self

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        res = handler(self)
        if "name" in res and res["name"] is None:
            res.pop("name")
        if "schemaId" in res and res["schemaId"] is None:
            res.pop("schemaId")
        if self.classifications:
            res["classifications"] = [
                c.model_dump(exclude_none=True) for c in self.classifications
            ]
        return res


class FrameLocation(BaseModel):
    end: int
    start: int


class VideoSupported(BaseModel):
    # Note that frames are only allowed as top level inferences for video
    frames: Optional[List[FrameLocation]] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        res = handler(self)
        # This means these are no video frames ..
        if self.frames is None:
            res.pop("frames")
        return res


class NDTextSubclass(NDAnswer):
    answer: str

    def to_common(self) -> Text:
        return Text(
            answer=self.answer,
            confidence=self.confidence,
            custom_metrics=self.custom_metrics,
        )

    @classmethod
    def from_common(
        cls, text: Text, name: str, feature_schema_id: Cuid
    ) -> "NDTextSubclass":
        return cls(
            answer=text.answer,
            name=name,
            schema_id=feature_schema_id,
            confidence=text.confidence,
            custom_metrics=text.custom_metrics,
        )


class NDChecklistSubclass(NDAnswer):
    answer: List[NDAnswer] = Field(..., validation_alias="answers")

    def to_common(self) -> Checklist:
        return Checklist(
            answer=[
                ClassificationAnswer(
                    name=answer.name,
                    feature_schema_id=answer.schema_id,
                    confidence=answer.confidence,
                    classifications=[
                        NDSubclassification.to_common(annot)
                        for annot in answer.classifications
                    ]
                    if answer.classifications
                    else None,
                    custom_metrics=answer.custom_metrics,
                )
                for answer in self.answer
            ]
        )

    @classmethod
    def from_common(
        cls, checklist: Checklist, name: str, feature_schema_id: Cuid
    ) -> "NDChecklistSubclass":
        return cls(
            answer=[
                NDAnswer(
                    name=answer.name,
                    schema_id=answer.feature_schema_id,
                    confidence=answer.confidence,
                    classifications=[
                        NDSubclassification.from_common(annot)
                        for annot in answer.classifications
                    ]
                    if answer.classifications
                    else None,
                    custom_metrics=answer.custom_metrics,
                )
                for answer in checklist.answer
            ],
            name=name,
            schema_id=feature_schema_id,
        )

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        res = handler(self)
        if "answers" in res:
            res["answer"] = res["answers"]
            del res["answers"]
        return res


class NDRadioSubclass(NDAnswer):
    answer: NDAnswer

    def to_common(self) -> Radio:
        return Radio(
            answer=ClassificationAnswer(
                name=self.answer.name,
                feature_schema_id=self.answer.schema_id,
                confidence=self.answer.confidence,
                classifications=[
                    NDSubclassification.to_common(annot)
                    for annot in self.answer.classifications
                ]
                if self.answer.classifications
                else None,
                custom_metrics=self.answer.custom_metrics,
            )
        )

    @classmethod
    def from_common(
        cls, radio: Radio, name: str, feature_schema_id: Cuid
    ) -> "NDRadioSubclass":
        return cls(
            answer=NDAnswer(
                name=radio.answer.name,
                schema_id=radio.answer.feature_schema_id,
                confidence=radio.answer.confidence,
                classifications=[
                    NDSubclassification.from_common(annot)
                    for annot in radio.answer.classifications
                ]
                if radio.answer.classifications
                else None,
                custom_metrics=radio.answer.custom_metrics,
            ),
            name=name,
            schema_id=feature_schema_id,
        )


class NDPromptTextSubclass(NDAnswer):
    answer: str

    def to_common(self) -> PromptText:
        return PromptText(
            answer=self.answer,
            confidence=self.confidence,
            custom_metrics=self.custom_metrics,
        )

    @classmethod
    def from_common(
        cls, prompt_text: PromptText, name: str, feature_schema_id: Cuid
    ) -> "NDPromptTextSubclass":
        return cls(
            answer=prompt_text.answer,
            name=name,
            schema_id=feature_schema_id,
            confidence=prompt_text.confidence,
            custom_metrics=prompt_text.custom_metrics,
        )


# ====== End of subclasses


class NDText(NDAnnotation, NDTextSubclass):
    @classmethod
    def from_common(
        cls,
        uuid: str,
        text: Text,
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        message_id: str,
        confidence: Optional[float] = None,
    ) -> "NDText":
        return cls(
            answer=text.answer,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            message_id=message_id,
            confidence=text.confidence,
            custom_metrics=text.custom_metrics,
        )


class NDChecklist(NDAnnotation, NDChecklistSubclass, VideoSupported):
    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        res = handler(self)
        if "classifications" in res and res["classifications"] == []:
            del res["classifications"]
        return res

    @classmethod
    def from_common(
        cls,
        uuid: str,
        checklist: Checklist,
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        message_id: str,
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
    ) -> "NDChecklist":
        return cls(
            answer=[
                NDAnswer(
                    name=answer.name,
                    schema_id=answer.feature_schema_id,
                    confidence=answer.confidence,
                    classifications=[
                        NDSubclassification.from_common(annot)
                        for annot in answer.classifications
                    ]
                    if answer.classifications
                    else None,
                    custom_metrics=answer.custom_metrics,
                )
                for answer in checklist.answer
            ],
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            frames=extra.get("frames"),
            message_id=message_id,
            confidence=confidence,
        )


class NDRadio(NDAnnotation, NDRadioSubclass, VideoSupported):
    @classmethod
    def from_common(
        cls,
        uuid: str,
        radio: Radio,
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: GenericDataRowData,
        message_id: str,
        confidence: Optional[float] = None,
    ) -> "NDRadio":
        return cls(
            answer=NDAnswer(
                name=radio.answer.name,
                schema_id=radio.answer.feature_schema_id,
                confidence=radio.answer.confidence,
                classifications=[
                    NDSubclassification.from_common(annot)
                    for annot in radio.answer.classifications
                ]
                if radio.answer.classifications
                else None,
                custom_metrics=radio.answer.custom_metrics,
            ),
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            frames=extra.get("frames"),
            message_id=message_id,
            confidence=confidence,
        )

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        res = handler(self)
        if "classifications" in res and res["classifications"] == []:
            del res["classifications"]
        return res


class NDPromptText(NDAnnotation, NDPromptTextSubclass):
    @classmethod
    def from_common(
        cls,
        uuid: str,
        text: PromptText,
        name,
        data: Dict,
        feature_schema_id: Cuid,
        confidence: Optional[float] = None,
    ) -> "NDPromptText":
        return cls(
            answer=text.answer,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            confidence=text.confidence,
            custom_metrics=text.custom_metrics,
        )


class NDSubclassification:
    @classmethod
    def from_common(
        cls, annotation: ClassificationAnnotation
    ) -> Union[NDTextSubclass, NDChecklistSubclass, NDRadioSubclass]:
        classify_obj = cls.lookup_subclassification(annotation)
        if classify_obj is None:
            raise TypeError(
                f"Unable to convert object to MAL format. `{type(annotation.value)}`"
            )
        return classify_obj.from_common(
            annotation.value, annotation.name, annotation.feature_schema_id
        )

    @staticmethod
    def to_common(
        annotation: "NDClassificationType",
    ) -> ClassificationAnnotation:
        return ClassificationAnnotation(
            value=annotation.to_common(),
            name=annotation.name,
            feature_schema_id=annotation.schema_id,
        )

    @staticmethod
    def lookup_subclassification(
        annotation: ClassificationAnnotation,
    ) -> Union[NDTextSubclass, NDChecklistSubclass, NDRadioSubclass]:
        return {
            Text: NDTextSubclass,
            Checklist: NDChecklistSubclass,
            Radio: NDRadioSubclass,
        }.get(type(annotation.value))


class NDClassification:
    @staticmethod
    def to_common(
        annotation: "NDClassificationType",
    ) -> Union[ClassificationAnnotation, VideoClassificationAnnotation]:
        common = ClassificationAnnotation(
            value=annotation.to_common(),
            name=annotation.name,
            feature_schema_id=annotation.schema_id,
            extra={"uuid": annotation.uuid},
            message_id=annotation.message_id,
            confidence=annotation.confidence,
        )

        if getattr(annotation, "frames", None) is None:
            return [common]
        results = []
        for frame in annotation.frames:
            for idx in range(frame.start, frame.end + 1, 1):
                results.append(
                    VideoClassificationAnnotation(
                        frame=idx, **common.model_dump(exclude_none=True)
                    )
                )
        return results

    @classmethod
    def from_common(
        cls,
        annotation: Union[
            ClassificationAnnotation, VideoClassificationAnnotation
        ],
        data: GenericDataRowData,
    ) -> Union[NDTextSubclass, NDChecklistSubclass, NDRadioSubclass]:
        classify_obj = cls.lookup_classification(annotation)
        if classify_obj is None:
            raise TypeError(
                f"Unable to convert object to MAL format. `{type(annotation.value)}`"
            )
        return classify_obj.from_common(
            str(annotation._uuid),
            annotation.value,
            annotation.name,
            annotation.feature_schema_id,
            annotation.extra,
            data,
            annotation.message_id,
            annotation.confidence,
        )

    @staticmethod
    def lookup_classification(
        annotation: Union[
            ClassificationAnnotation, VideoClassificationAnnotation
        ],
    ) -> Union[NDText, NDChecklist, NDRadio]:
        return {Text: NDText, Checklist: NDChecklist, Radio: NDRadio}.get(
            type(annotation.value)
        )


class NDPromptClassification:
    @staticmethod
    def to_common(
        annotation: "NDPromptClassificationType",
    ) -> Union[PromptClassificationAnnotation]:
        common = PromptClassificationAnnotation(
            value=annotation,
            name=annotation.name,
            feature_schema_id=annotation.schema_id,
            extra={"uuid": annotation.uuid},
            confidence=annotation.confidence,
        )

        return common

    @classmethod
    def from_common(
        cls,
        annotation: Union[PromptClassificationAnnotation],
        data: GenericDataRowData,
    ) -> Union[NDTextSubclass, NDChecklistSubclass, NDRadioSubclass]:
        return NDPromptText.from_common(
            str(annotation._uuid),
            annotation.value,
            annotation.name,
            data,
            annotation.feature_schema_id,
            annotation.confidence,
        )


# Make sure to keep NDChecklistSubclass prior to NDRadioSubclass in the list,
# otherwise list of answers gets parsed by NDRadio whereas NDChecklist must be used
NDSubclassificationType = Union[
    NDChecklistSubclass, NDRadioSubclass, NDTextSubclass
]

NDAnswer.model_rebuild()
NDChecklistSubclass.model_rebuild()
NDChecklist.model_rebuild()
NDRadioSubclass.model_rebuild()
NDRadio.model_rebuild()
NDText.model_rebuild()
NDPromptText.model_rebuild()
NDTextSubclass.model_rebuild()

# Make sure to keep NDChecklist prior to NDRadio in the list,
# otherwise list of answers gets parsed by NDRadio whereas NDChecklist must be used
NDClassificationType = Union[NDChecklist, NDRadio, NDText]
NDPromptClassificationType = Union[NDPromptText]



from datetime import datetime, timezone
from typing import Collection, Dict, List, Literal, Optional, Tuple, TypedDict

SEARCH_LIMIT_PER_EXPORT_V2 = 2_000
ISO_8061_FORMAT = "%Y-%m-%dT%H:%M:%S%z"


class BaseExportFilters(TypedDict):
    data_row_ids: Optional[List[str]]
    """ Datarow ids to export
    Please refer to https://docs.labelbox.com/docs/limits#export on the allowed limit of data_row_ids
    Example:
    >>>   ["clgo3lyax0000veeezdbu3ws4", "clgo3lzjl0001veeer6y6z8zp", ...]

    """

    global_keys: Optional[List[str]]
    """ Global keys to export
    Please refer to https://docs.labelbox.com/docs/limits#export on the allowed limit of data_row_ids
    Example:
    >>>   ["key1", "key2", ...]
    """


class SharedExportFilters(BaseExportFilters):
    label_created_at: Optional[Tuple[str, str]]
    """ Date range for labels created at
    Formatted "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss"
    Examples: 
    >>>   ["2000-01-01 00:00:00", "2050-01-01 00:00:00"]
    >>>   [None, "2050-01-01 00:00:00"]
    >>>   ["2000-01-01 00:00:00", None]
    """
    last_activity_at: Optional[Tuple[str, str]]


class ProjectExportFilters(SharedExportFilters):
    batch_ids: Optional[List[str]]
    """ Batch ids to export
    Example:
    >>> ["clgo3lyax0000veeezdbu3ws4"]
    """
    workflow_status: Optional[
        Literal["ToLabel", "InReview", "InRework", "Done"]
    ]
    """ Export data rows matching workflow status
    Example:
    >>> "InReview"
    """


class DatasetExportFilters(SharedExportFilters):
    pass


class CatalogExportFilters(SharedExportFilters):
    pass


class DatarowExportFilters(BaseExportFilters):
    pass


def validate_datetime(datetime_str: str) -> bool:
    """helper function to validate that datetime's format: "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss"
    or ISO 8061 format "YYYY-MM-DDThh:mm:ss±hhmm" (Example: "2023-05-23T14:30:00+0530")"""
    if datetime_str:
        for fmt in ("%Y-%m-%d", "%Y-%m-%d %H:%M:%S", ISO_8061_FORMAT):
            try:
                datetime.strptime(datetime_str, fmt)
                return True
            except ValueError:
                pass
        raise ValueError(f"""Incorrect format for: {datetime_str}.
        Format must be \"YYYY-MM-DD\" or \"YYYY-MM-DD hh:mm:ss\" or ISO 8061 format \"YYYY-MM-DDThh:mm:ss±hhmm\"""")
    return True


def convert_to_utc_if_iso8061(datetime_str: str, timezone_str: Optional[str]):
    """helper function to convert datetime to UTC if it is in ISO_8061_FORMAT and set timezone to UTC"""
    try:
        date_obj = datetime.strptime(datetime_str, ISO_8061_FORMAT)
        date_obj_utc = date_obj.replace(tzinfo=timezone.utc)
        datetime_str = date_obj_utc.strftime(ISO_8061_FORMAT)
        timezone_str = "UTC"
    except ValueError:
        pass
    return datetime_str, timezone_str


def validate_one_of_data_row_ids_or_global_keys(filters):
    if (
        filters.get("data_row_ids") is not None
        and filters.get("global_keys") is not None
    ):
        raise ValueError(
            "data_rows and global_keys cannot both be present in export filters"
        )


def validate_at_least_one_of_data_row_ids_or_global_keys(filters):
    if not filters.get("data_row_ids") and not filters.get("global_keys"):
        raise ValueError("data_rows and global_keys cannot both be empty")


def build_filters(client, filters):
    search_query: List[Dict[str, Collection[str]]] = []
    timezone: Optional[str] = None

    def _get_timezone() -> str:
        timezone_query_str = """query CurrentUserPyApi { user { timezone } }"""
        tz_res = client.execute(timezone_query_str)
        return tz_res["user"]["timezone"] or "UTC"

    def _build_id_filters(
        ids: list,
        type_name: str,
        search_where_limit: int = SEARCH_LIMIT_PER_EXPORT_V2,
    ):
        if not isinstance(ids, list):
            raise ValueError(f"{type_name} filter expects a list.")
        if len(ids) == 0:
            raise ValueError(f"{type_name} filter expects a non-empty list.")
        if len(ids) > search_where_limit:
            raise ValueError(
                f"{type_name} filter only supports a max of {search_where_limit} items."
            )
        search_query.append({"ids": ids, "operator": "is", "type": type_name})

    validate_one_of_data_row_ids_or_global_keys(filters)

    last_activity_at = filters.get("last_activity_at")
    if last_activity_at:
        timezone = _get_timezone()
        start, end = last_activity_at
        if start is not None and end is not None:
            [validate_datetime(date) for date in last_activity_at]
            start, timezone = convert_to_utc_if_iso8061(start, timezone)
            end, timezone = convert_to_utc_if_iso8061(end, timezone)
            search_query.append(
                {
                    "type": "data_row_last_activity_at",
                    "value": {
                        "operator": "BETWEEN",
                        "timezone": timezone,
                        "value": {"min": start, "max": end},
                    },
                }
            )
        elif start is not None:
            validate_datetime(start)
            start, timezone = convert_to_utc_if_iso8061(start, timezone)
            search_query.append(
                {
                    "type": "data_row_last_activity_at",
                    "value": {
                        "operator": "GREATER_THAN_OR_EQUAL",
                        "timezone": timezone,
                        "value": start,
                    },
                }
            )
        elif end is not None:
            validate_datetime(end)
            end, timezone = convert_to_utc_if_iso8061(end, timezone)
            search_query.append(
                {
                    "type": "data_row_last_activity_at",
                    "value": {
                        "operator": "LESS_THAN_OR_EQUAL",
                        "timezone": timezone,
                        "value": end,
                    },
                }
            )

    label_created_at = filters.get("label_created_at")
    if label_created_at:
        timezone = _get_timezone()
        start, end = label_created_at
        if start is not None and end is not None:
            [validate_datetime(date) for date in label_created_at]
            start, timezone = convert_to_utc_if_iso8061(start, timezone)
            end, timezone = convert_to_utc_if_iso8061(end, timezone)
            search_query.append(
                {
                    "type": "labeled_at",
                    "value": {
                        "operator": "BETWEEN",
                        "timezone": timezone,
                        "value": {"min": start, "max": end},
                    },
                }
            )
        elif start is not None:
            validate_datetime(start)
            start, timezone = convert_to_utc_if_iso8061(start, timezone)
            search_query.append(
                {
                    "type": "labeled_at",
                    "value": {
                        "operator": "GREATER_THAN_OR_EQUAL",
                        "timezone": timezone,
                        "value": start,
                    },
                }
            )
        elif end is not None:
            validate_datetime(end)
            end, timezone = convert_to_utc_if_iso8061(end, timezone)
            search_query.append(
                {
                    "type": "labeled_at",
                    "value": {
                        "operator": "LESS_THAN_OR_EQUAL",
                        "timezone": timezone,
                        "value": end,
                    },
                }
            )

    data_row_ids = filters.get("data_row_ids")
    if data_row_ids is not None:
        _build_id_filters(data_row_ids, "data_row_id")

    global_keys = filters.get("global_keys")
    if global_keys is not None:
        _build_id_filters(global_keys, "global_key")

    batch_ids = filters.get("batch_ids")
    if batch_ids is not None:
        _build_id_filters(batch_ids, "batch")

    workflow_status = filters.get("workflow_status")
    if workflow_status:
        if not isinstance(workflow_status, str):
            raise ValueError("`workflow_status` filter expects a string.")
        elif workflow_status not in ["ToLabel", "InReview", "InRework", "Done"]:
            raise ValueError(
                "`workflow_status` filter expects one of 'InReview', 'InRework', or 'Done'."
            )

        if workflow_status == "ToLabel":
            search_query.append({"type": "task_queue_not_exist"})
        else:
            search_query.append(
                {"type": "task_queue_status", "status": workflow_status}
            )

    return search_query



from abc import ABC
from typing import Optional, Dict, List, Any

from pydantic import BaseModel


class BaseData(BaseModel, ABC):
    """
    Base class for objects representing data.
    This class shouldn't directly be used
    """

    external_id: Optional[str] = None
    uid: Optional[str] = None
    global_key: Optional[str] = None
    media_attributes: Optional[Dict[str, Any]] = None
    metadata: Optional[List[Dict[str, Any]]] = None



from unittest.mock import MagicMock

import pytest

from labelbox.schema.user_group_v2 import Member, UserGroupV2


@pytest.fixture
def client():
    return MagicMock()


def test_parse_members_csv_empty(client):
    group = UserGroupV2(client)
    assert group._parse_members_csv("") == []
    assert group._parse_members_csv("\n") == []


def test_parse_members_csv_header_only(client):
    group = UserGroupV2(client)
    assert group._parse_members_csv("email\n") == []


def test_parse_members_csv_single_member(client):
    group = UserGroupV2(client)
    result = group._parse_members_csv("email\ntest@example.com")
    assert len(result) == 1
    assert isinstance(result[0], Member)
    assert result[0].email == "test@example.com"


def test_parse_members_csv_multiple_members(client):
    group = UserGroupV2(client)
    csv_data = "email\ntest1@example.com\ntest2@example.com\ntest3@example.com"
    result = group._parse_members_csv(csv_data)
    assert len(result) == 3
    assert [m.email for m in result] == [
        "test1@example.com",
        "test2@example.com",
        "test3@example.com",
    ]


def test_parse_members_csv_handles_whitespace(client):
    group = UserGroupV2(client)
    csv_data = "email\n  test@example.com  \n\nother@example.com\n"
    result = group._parse_members_csv(csv_data)
    assert len(result) == 2
    assert result[0].email == "test@example.com"
    assert result[1].email == "other@example.com"



from types import SimpleNamespace
from uuid import uuid4

import numpy as np
import pytest

from labelbox.data.annotation_types import (
    LabelGenerator,
    ObjectAnnotation,
    Mask,
    Label,
    GenericDataRowData,
    MaskData,
)


@pytest.fixture
def list_of_labels():
    return [
        Label(data=GenericDataRowData(uid="http://someurl")) for _ in range(5)
    ]


@pytest.fixture
def signer():
    def get_signer(uuid):
        return lambda x: uuid

    return get_signer


class FakeDataset:
    def __init__(self):
        self.uid = "ckrb4tgm51xl10ybc7lv9ghm7"
        self.exports = []

    def create_data_row(self, row_data, external_id=None):
        if external_id is None:
            external_id = "an external_id"
        return SimpleNamespace(uid=self.uid, external_id=external_id)

    def create_data_rows(self, args):
        for arg in args:
            self.exports.append(
                SimpleNamespace(
                    row_data=arg["row_data"],
                    external_id=arg["external_id"],
                    uid=self.uid,
                )
            )
        return self

    def wait_till_done(self):
        pass


def test_generator(list_of_labels):
    generator = LabelGenerator([list_of_labels[0]])

    assert next(generator) == list_of_labels[0]
    with pytest.raises(StopIteration):
        next(generator)


def test_conversion(list_of_labels):
    generator = LabelGenerator(list_of_labels)
    label_collection = list(generator)
    assert len(label_collection) == len(list_of_labels)
    assert [x for x in label_collection] == list_of_labels


def test_adding_to_masks(signer):
    label = Label(
        data=GenericDataRowData(uid="12345"),
        annotations=[
            ObjectAnnotation(
                name="1234",
                value=Mask(
                    mask=MaskData(
                        arr=np.random.random((32, 32, 3)).astype(np.uint8)
                    ),
                    color=[255, 255, 255],
                ),
            )
        ],
    )
    uuid = str(uuid4())
    generator = LabelGenerator([label]).add_url_to_masks(signer(uuid))
    assert label.annotations[0].value.mask.url != uuid
    assert next(generator).annotations[0].value.mask.url == uuid
    assert label.annotations[0].value.mask.url == uuid



import uuid
from labelbox import parser
from labelbox.data.annotation_types.annotation import ObjectAnnotation
from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.annotation_types.geometry.line import Line
from labelbox.data.annotation_types.geometry.point import Point
from labelbox.data.annotation_types.geometry.polygon import Polygon
from labelbox.data.annotation_types.geometry.rectangle import Rectangle
from labelbox.data.annotation_types.label import Label
import pytest
from labelbox import ModelRun

from labelbox.schema.annotation_import import (
    AnnotationImportState,
    MEAPredictionImport,
)
from labelbox.data.serialization import NDJsonConverter
from labelbox.schema.export_params import ModelRunExportParams

"""
- Here we only want to check that the uploads are calling the validation
- Then with unit tests we can check the types of errors raised

"""


@pytest.mark.order(1)
def test_create_from_objects(
    model_run_with_data_rows,
    object_predictions_for_annotation_import,
    annotation_import_test_helpers,
):
    name = str(uuid.uuid4())
    object_predictions = object_predictions_for_annotation_import
    use_data_row_ids = [p["dataRow"]["id"] for p in object_predictions]
    model_run_with_data_rows.upsert_data_rows(use_data_row_ids)

    annotation_import = model_run_with_data_rows.add_predictions(
        name=name, predictions=object_predictions
    )

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import_test_helpers.check_running_state(annotation_import, name)
    annotation_import_test_helpers.assert_file_content(
        annotation_import.input_file_url, object_predictions
    )
    annotation_import.wait_until_done()

    assert annotation_import.state == AnnotationImportState.FINISHED
    annotation_import_test_helpers.download_and_assert_status(
        annotation_import.status_file_url
    )


def test_create_from_objects_global_key(
    client,
    model_run_with_data_rows,
    polygon_inference,
    annotation_import_test_helpers,
):
    name = str(uuid.uuid4())
    dr = client.get_data_row(polygon_inference[0]["dataRow"]["id"])
    polygon_inference[0]["dataRow"]["globalKey"] = dr.global_key
    del polygon_inference[0]["dataRow"]["id"]
    object_predictions = [polygon_inference[0]]

    annotation_import = model_run_with_data_rows.add_predictions(
        name=name, predictions=object_predictions
    )

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import_test_helpers.check_running_state(annotation_import, name)
    annotation_import_test_helpers.assert_file_content(
        annotation_import.input_file_url, object_predictions
    )
    annotation_import.wait_until_done()

    assert annotation_import.state == AnnotationImportState.FINISHED
    annotation_import_test_helpers.download_and_assert_status(
        annotation_import.status_file_url
    )


def test_create_from_objects_with_confidence(
    predictions_with_confidence,
    model_run_with_data_rows,
    annotation_import_test_helpers,
):
    name = str(uuid.uuid4())

    object_prediction_data_rows = set(
        [
            object_prediction["dataRow"]["id"]
            for object_prediction in predictions_with_confidence
        ]
    )
    # MUST have all data rows in the model run
    model_run_with_data_rows.upsert_data_rows(
        data_row_ids=list(object_prediction_data_rows)
    )

    annotation_import = model_run_with_data_rows.add_predictions(
        name=name, predictions=predictions_with_confidence
    )

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import_test_helpers.check_running_state(annotation_import, name)
    annotation_import_test_helpers.assert_file_content(
        annotation_import.input_file_url, predictions_with_confidence
    )
    annotation_import.wait_until_done()

    assert annotation_import.state == AnnotationImportState.FINISHED
    annotation_import_test_helpers.download_and_assert_status(
        annotation_import.status_file_url
    )


def test_create_from_objects_all_project_labels(
    model_run_with_all_project_labels,
    object_predictions_for_annotation_import,
    annotation_import_test_helpers,
):
    name = str(uuid.uuid4())
    object_predictions = object_predictions_for_annotation_import
    use_data_row_ids = [p["dataRow"]["id"] for p in object_predictions]
    model_run_with_all_project_labels.upsert_data_rows(use_data_row_ids)

    annotation_import = model_run_with_all_project_labels.add_predictions(
        name=name, predictions=object_predictions
    )

    assert (
        annotation_import.model_run_id == model_run_with_all_project_labels.uid
    )
    annotation_import_test_helpers.check_running_state(annotation_import, name)
    annotation_import_test_helpers.assert_file_content(
        annotation_import.input_file_url, object_predictions
    )
    annotation_import.wait_until_done()

    assert annotation_import.state == AnnotationImportState.FINISHED
    annotation_import_test_helpers.download_and_assert_status(
        annotation_import.status_file_url
    )


def test_model_run_project_labels(
    model_run_with_all_project_labels: ModelRun, model_run_predictions
):
    model_run = model_run_with_all_project_labels
    export_task = model_run.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()

    # exports to list of tuples (data_row_id, label) needed to adapt test to export v2 instead of export v1 since data rows ids are not at label level in export v2.
    model_run_exported_labels = [
        (
            data_row.json["data_row"]["id"],
            data_row.json["experiments"][model_run.model_id]["runs"][
                model_run.uid
            ]["labels"][0],
        )
        for data_row in stream
    ]

    labels_indexed_by_name = {}

    # making sure the labels are in this model run are all labels uploaded to the project
    # by comparing some 'immutable' attributes
    # multiple data rows per prediction import
    for data_row_id, label in model_run_exported_labels:
        for object in label["annotations"]["objects"]:
            name = object["name"]
            labels_indexed_by_name[f"{name}-{data_row_id}"] = {
                "label": label,
                "data_row_id": data_row_id,
            }

    assert (len(labels_indexed_by_name.keys())) == len(
        [prediction["dataRow"]["id"] for prediction in model_run_predictions]
    )

    expected_data_row_ids = set(
        [prediction["dataRow"]["id"] for prediction in model_run_predictions]
    )
    expected_objects = set(
        [prediction["name"] for prediction in model_run_predictions]
    )
    for data_row_id, actual_label in model_run_exported_labels:
        assert data_row_id in expected_data_row_ids
        assert len(expected_objects) == len(
            actual_label["annotations"]["objects"]
        )


def test_create_from_label_objects(
    model_run_with_data_rows,
    object_predictions_for_annotation_import,
    annotation_import_test_helpers,
):
    name = str(uuid.uuid4())
    use_data_row_id = [
        p["dataRow"]["id"] for p in object_predictions_for_annotation_import
    ]

    model_run_with_data_rows.upsert_data_rows(use_data_row_id)

    predictions = []
    for data_row_id in use_data_row_id:
        predictions.append(
            Label(
                data=GenericDataRowData(
                    uid=data_row_id,
                ),
                annotations=[
                    ObjectAnnotation(
                        name="polygon",
                        value=Polygon(
                            points=[
                                Point(x=147.692, y=118.154),
                                Point(x=142.769, y=104.923),
                                Point(x=57.846, y=118.769),
                                Point(x=28.308, y=169.846),
                                Point(x=147.692, y=118.154),
                            ],
                        ),
                    ),
                    ObjectAnnotation(
                        name="bbox",
                        value=Rectangle(
                            start=Point(x=58.0, y=48.0),
                            end=Point(x=70.0, y=113.0),
                        ),
                    ),
                    ObjectAnnotation(
                        name="polyline",
                        value=Line(
                            points=[
                                Point(x=147.692, y=118.154),
                                Point(x=150.692, y=160.154),
                            ],
                        ),
                    ),
                ],
            ),
        )

    annotation_import = model_run_with_data_rows.add_predictions(
        name=name, predictions=predictions
    )

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import_test_helpers.check_running_state(annotation_import, name)
    normalized_predictions = NDJsonConverter.serialize(predictions)
    annotation_import_test_helpers.assert_file_content(
        annotation_import.input_file_url, normalized_predictions
    )
    annotation_import.wait_until_done()

    assert annotation_import.state == AnnotationImportState.FINISHED
    annotation_import_test_helpers.download_and_assert_status(
        annotation_import.status_file_url
    )


def test_create_from_local_file(
    tmp_path,
    model_run_with_data_rows,
    object_predictions_for_annotation_import,
    annotation_import_test_helpers,
):
    use_data_row_ids = [
        p["dataRow"]["id"] for p in object_predictions_for_annotation_import
    ]
    model_run_with_data_rows.upsert_data_rows(use_data_row_ids)

    name = str(uuid.uuid4())
    file_name = f"{name}.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        parser.dump(object_predictions_for_annotation_import, f)

    annotation_import = model_run_with_data_rows.add_predictions(
        name=name, predictions=str(file_path)
    )

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import_test_helpers.check_running_state(annotation_import, name)
    annotation_import_test_helpers.assert_file_content(
        annotation_import.input_file_url,
        object_predictions_for_annotation_import,
    )
    annotation_import.wait_until_done()

    assert annotation_import.state == AnnotationImportState.FINISHED
    annotation_import_test_helpers.download_and_assert_status(
        annotation_import.status_file_url
    )


def test_predictions_with_custom_metrics(
    model_run,
    object_predictions_for_annotation_import,
    annotation_import_test_helpers,
):
    name = str(uuid.uuid4())
    object_predictions = object_predictions_for_annotation_import
    use_data_row_ids = [p["dataRow"]["id"] for p in object_predictions]
    model_run.upsert_data_rows(use_data_row_ids)

    annotation_import = model_run.add_predictions(
        name=name, predictions=object_predictions
    )

    assert annotation_import.model_run_id == model_run.uid
    annotation_import.wait_until_done()
    assert annotation_import.state == AnnotationImportState.FINISHED

    task = model_run.export_v2(params=ModelRunExportParams(predictions=True))
    task.wait_till_done()

    assert annotation_import.state == AnnotationImportState.FINISHED
    annotation_import_test_helpers.download_and_assert_status(
        annotation_import.status_file_url
    )


def test_get(client, model_run_with_data_rows, annotation_import_test_helpers):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    model_run_with_data_rows.add_predictions(name=name, predictions=url)

    annotation_import = MEAPredictionImport.from_name(
        client, model_run_id=model_run_with_data_rows.uid, name=name
    )

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import_test_helpers.check_running_state(
        annotation_import, name, url
    )
    annotation_import.wait_until_done()


@pytest.mark.slow
def test_wait_till_done(model_run_predictions, model_run_with_data_rows):
    name = str(uuid.uuid4())
    annotation_import = model_run_with_data_rows.add_predictions(
        name=name, predictions=model_run_predictions
    )

    assert len(annotation_import.inputs) == len(model_run_predictions)
    annotation_import.wait_until_done()
    assert annotation_import.state == AnnotationImportState.FINISHED
    # Check that the status files are being returned as expected
    assert len(annotation_import.errors) == 0
    assert len(annotation_import.inputs) == len(model_run_predictions)
    input_uuids = [
        input_annot["uuid"] for input_annot in annotation_import.inputs
    ]
    inference_uuids = [pred["uuid"] for pred in model_run_predictions]
    assert set(input_uuids) == set(inference_uuids)
    assert len(annotation_import.statuses) == len(model_run_predictions)
    for status in annotation_import.statuses:
        assert status["status"] == "SUCCESS"
    status_uuids = [
        input_annot["uuid"] for input_annot in annotation_import.statuses
    ]
    assert set(input_uuids) == set(status_uuids)



def test_benchmark(configured_project_with_label):
    project, _, data_row, label = configured_project_with_label
    assert set(project.benchmarks()) == set()
    assert label.is_benchmark_reference is False

    benchmark = label.create_benchmark()
    assert set(project.benchmarks()) == {benchmark}
    assert benchmark.reference_label() == label
    # Refresh label data to check it's benchmark reference
    label = list(data_row.labels())[0]
    assert label.is_benchmark_reference is True

    benchmark.delete()
    assert set(project.benchmarks()) == set()
    # Refresh label data to check it's benchmark reference
    label = list(data_row.labels())[0]
    assert label.is_benchmark_reference is False



import labelbox.types as lb_types


def test_mask_frame():
    mask_frame = lb_types.MaskFrame(
        index=1, instance_uri="http://path/to/frame.png"
    )
    assert mask_frame.model_dump(by_alias=True) == {
        "index": 1,
        "imBytes": None,
        "instanceURI": "http://path/to/frame.png",
    }


def test_mask_instance():
    mask_instance = lb_types.MaskInstance(color_rgb=(0, 0, 255), name="mask1")
    assert mask_instance.model_dump(by_alias=True, exclude_none=True) == {
        "colorRGB": (0, 0, 255),
        "name": "mask1",
    }



import json
import pytest


@pytest.fixture
def ndjson_content():
    line = """{"uuid": "9fd9a92e-2560-4e77-81d4-b2e955800092", "schemaId": "ckaeasyfk004y0y7wyye5epgu", "dataRow": {"id": "ck7kftpan8ir008910yf07r9c"}, "bbox": {"top": 48, "left": 58, "height": 865, "width": 1512}}
{"uuid": "29b878f3-c2b4-4dbf-9f22-a795f0720125", "schemaId": "ckapgvrl7007q0y7ujkjkaaxt", "dataRow": {"id": "ck7kftpan8ir008910yf07r9c"}, "polygon": [{"x": 147.692, "y": 118.154}, {"x": 142.769, "y": 404.923}, {"x": 57.846, "y": 318.769}, {"x": 28.308, "y": 169.846}]}"""
    expected_objects = [
        {
            "uuid": "9fd9a92e-2560-4e77-81d4-b2e955800092",
            "schemaId": "ckaeasyfk004y0y7wyye5epgu",
            "dataRow": {"id": "ck7kftpan8ir008910yf07r9c"},
            "bbox": {"top": 48, "left": 58, "height": 865, "width": 1512},
        },
        {
            "uuid": "29b878f3-c2b4-4dbf-9f22-a795f0720125",
            "schemaId": "ckapgvrl7007q0y7ujkjkaaxt",
            "dataRow": {"id": "ck7kftpan8ir008910yf07r9c"},
            "polygon": [
                {"x": 147.692, "y": 118.154},
                {"x": 142.769, "y": 404.923},
                {"x": 57.846, "y": 318.769},
                {"x": 28.308, "y": 169.846},
            ],
        },
    ]

    return line, expected_objects


@pytest.fixture
def ndjson_content_with_nonascii_and_line_breaks():
    line = '{"id": "2489651127", "type": "PushEvent", "actor": {"id": 1459915, "login": "xtuaok", "gravatar_id": "", "url": "https://api.github.com/users/xtuaok", "avatar_url": "https://avatars.githubusercontent.com/u/1459915?"}, "repo": {"id": 6719841, "name": "xtuaok/twitter_track_following", "url": "https://api.github.com/repos/xtuaok/twitter_track_following"}, "payload": {"push_id": 536864008, "size": 1, "distinct_size": 1, "ref": "refs/heads/xtuaok", "head": "afb8afe306c7893d93d383a06e4d9df53b41bf47", "before": "4671b4868f1a060f2ed64d8268cd22d514a84e63", "commits": [{"sha": "afb8afe306c7893d93d383a06e4d9df53b41bf47", "author": {"email": "47cb89439b2d6961b59dff4298e837f67aa77389@gmail.com", "name": "Tomonori Tamagawa"}, "message": "Update ID 949438177,, - screen_name: chomado, - name: ちょまど@初詣おみくじ凶, - description: ( *ﾟ▽ﾟ* っ)З腐女子！絵描き！| H26新卒文系SE (入社して4ヶ月目の8月にSIer(適応障害になった)を辞職し開発者に転職) | H26秋応用情報合格！| 自作bot (in PHP) chomado_bot | プログラミングガチ初心者, - location:", "distinct": true, "url": "https://api.github.com/repos/xtuaok/twitter_track_following/commits/afb8afe306c7893d93d383a06e4d9df53b41bf47"}]}, "public": true, "created_at": "2015-01-01T15:00:10Z"}'
    expected_objects = [
        {
            "id": "2489651127",
            "type": "PushEvent",
            "actor": {
                "id": 1459915,
                "login": "xtuaok",
                "gravatar_id": "",
                "url": "https://api.github.com/users/xtuaok",
                "avatar_url": "https://avatars.githubusercontent.com/u/1459915?",
            },
            "repo": {
                "id": 6719841,
                "name": "xtuaok/twitter_track_following",
                "url": "https://api.github.com/repos/xtuaok/twitter_track_following",
            },
            "payload": {
                "push_id": 536864008,
                "size": 1,
                "distinct_size": 1,
                "ref": "refs/heads/xtuaok",
                "head": "afb8afe306c7893d93d383a06e4d9df53b41bf47",
                "before": "4671b4868f1a060f2ed64d8268cd22d514a84e63",
                "commits": [
                    {
                        "sha": "afb8afe306c7893d93d383a06e4d9df53b41bf47",
                        "author": {
                            "email": "47cb89439b2d6961b59dff4298e837f67aa77389@gmail.com",
                            "name": "Tomonori Tamagawa",
                        },
                        "message": "Update ID 949438177,, - screen_name: chomado, - name: ちょまど@初詣おみくじ凶, - description: ( *ﾟ▽ﾟ* っ)З腐女子！絵描き！| H26新卒文系SE (入社して4ヶ月目の8月にSIer(適応障害になった)を辞職し開発者に転職) | H26秋応用情報合格！| 自作bot (in PHP) chomado_bot | プログラミングガチ初心者, - location:",
                        "distinct": True,
                        "url": "https://api.github.com/repos/xtuaok/twitter_track_following/commits/afb8afe306c7893d93d383a06e4d9df53b41bf47",
                    }
                ],
            },
            "public": True,
            "created_at": "2015-01-01T15:00:10Z",
        }
    ]
    return line, expected_objects


@pytest.fixture
def generate_random_ndjson(rand_gen):
    def _generate_random_ndjson(lines: int = 10):
        return [
            json.dumps({"data_row": {"id": rand_gen(str)}})
            for _ in range(lines)
        ]

    return _generate_random_ndjson


@pytest.fixture
def mock_response():
    class MockResponse:
        def __init__(self, text: str, exception: Exception = None) -> None:
            self._text = text
            self._exception = exception

        @property
        def text(self):
            return self._text

        @property
        def content(self):
            return self._text

        def raise_for_status(self):
            if self._exception:
                raise self._exception

    return MockResponse



import warnings
from abc import ABC
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set

from labelbox.schema.tool_building.tool_type import ToolType


@dataclass
class _Variant:
    id: int
    name: str
    actions: List[str] = field(default_factory=list)
    _available_actions: Set[str] = field(default_factory=set)

    def set_actions(self, actions: List[str]) -> None:
        self.actions = []
        for action in actions:
            if action in self._available_actions:
                self.actions.append(action)
            else:
                warnings.warn(
                    f"Variant ID {self.id} {action} is an invalid action, skipping"
                )

    def reset_actions(self) -> None:
        self.actions = []

    def asdict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "actions": self.actions,
        }

    def _post_init(self):
        # Call set_actions to remove any invalid actions
        self.set_actions(self.actions)


@dataclass
class _Definition:
    variants: List[_Variant]
    version: int = field(default=1)
    title: Optional[str] = None
    value: Optional[str] = None

    def __post_init__(self):
        if self.version != 1:
            raise ValueError("Invalid version")

    def asdict(self) -> Dict[str, Any]:
        result = {
            "variants": [variant.asdict() for variant in self.variants],
            "version": self.version,
        }
        if self.title is not None:
            result["title"] = self.title
        if self.value is not None:
            result["value"] = self.value
        return result

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "_Definition":
        variants = [_Variant(**variant) for variant in dictionary["variants"]]
        title = dictionary.get("title", None)
        value = dictionary.get("value", None)
        return cls(variants=variants, title=title, value=value)


@dataclass
class _BaseStepReasoningTool(ABC):
    name: str
    definition: _Definition
    type: Optional[ToolType] = None
    schema_id: Optional[str] = None
    feature_schema_id: Optional[str] = None
    color: Optional[str] = None
    required: bool = False  # This attribute is for consistency with other tools and backend, default is False

    def __post_init__(self):
        if not self.name.strip():
            raise ValueError("Name is required")

    def asdict(self) -> Dict[str, Any]:
        return {
            "tool": self.type.value if self.type else None,
            "name": self.name,
            "required": self.required,
            "schemaNodeId": self.schema_id,
            "featureSchemaId": self.feature_schema_id,
            "definition": self.definition.asdict(),
            "color": self.color,
        }

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "_BaseStepReasoningTool":
        return cls(
            name=dictionary["name"],
            schema_id=dictionary.get("schemaNodeId", None),
            feature_schema_id=dictionary.get("featureSchemaId", None),
            required=dictionary.get("required", False),
            definition=_Definition.from_dict(dictionary["definition"]),
            color=dictionary.get("color", None),
        )



from typing import Union

from labelbox.utils import _CamelCaseMixin

from .base import DataRow, NDAnnotation
from ...annotation_types.mmc import (
    MessageSingleSelectionTask,
    MessageMultiSelectionTask,
    MessageRankingTask,
    MessageEvaluationTaskAnnotation,
)
from ...annotation_types import GenericDataRowData


class MessageTaskData(_CamelCaseMixin):
    format: str
    data: Union[
        MessageSingleSelectionTask,
        MessageMultiSelectionTask,
        MessageRankingTask,
    ]


class NDMessageTask(NDAnnotation):
    message_evaluation_task: MessageTaskData

    def to_common(self) -> MessageEvaluationTaskAnnotation:
        return MessageEvaluationTaskAnnotation(
            name=self.name,
            feature_schema_id=self.schema_id,
            value=self.message_evaluation_task.data,
            extra={"uuid": self.uuid},
        )

    @classmethod
    def from_common(
        cls,
        annotation: MessageEvaluationTaskAnnotation,
        data: GenericDataRowData,
    ) -> "NDMessageTask":
        return cls(
            uuid=str(annotation._uuid),
            name=annotation.name,
            schema_id=annotation.feature_schema_id,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            message_evaluation_task=MessageTaskData(
                format=annotation.value.format, data=annotation.value
            ),
        )



import numpy as np

import labelbox.types as lb_types
from labelbox.data.annotation_types import (
    PromptText,
    ObjectAnnotation,
    Point,
    Line,
    MaskData,
    Label,
)
import pytest


def test_schema_assignment_confidence():
    name = "line_feature"
    label = Label(
        data=MaskData(arr=np.ones((32, 32, 3), dtype=np.uint8), uid="test"),
        annotations=[
            ObjectAnnotation(
                value=Line(
                    points=[Point(x=1, y=2), Point(x=2, y=2)],
                ),
                name=name,
                confidence=0.914,
            )
        ],
    )

    assert label.annotations[0].confidence == 0.914


def test_initialize_label_no_coercion():
    global_key = "global-key"
    ner_annotation = lb_types.ObjectAnnotation(
        name="ner",
        value=lb_types.ConversationEntity(start=0, end=8, message_id="4"),
    )
    label = Label(
        data=lb_types.GenericDataRowData(global_key=global_key),
        annotations=[ner_annotation],
    )
    assert isinstance(label.data, lb_types.GenericDataRowData)
    assert label.data.global_key == global_key


def test_prompt_classification_validation():
    global_key = "global-key"
    prompt_text = lb_types.PromptClassificationAnnotation(
        name="prompt text", value=PromptText(answer="test")
    )
    prompt_text_2 = lb_types.PromptClassificationAnnotation(
        name="prompt text", value=PromptText(answer="test")
    )
    with pytest.raises(TypeError) as e_info:
        label = Label(
            data={"global_key": global_key},
            annotations=[prompt_text, prompt_text_2],
        )



from typing import Annotated, Optional, Set

from pydantic import BaseModel, ConfigDict, Field, model_validator

from labelbox.schema.media_type import MediaType
from labelbox.schema.ontology_kind import EditorTaskType
from labelbox.schema.quality_mode import (
    BENCHMARK_AUTO_AUDIT_NUMBER_OF_LABELS,
    BENCHMARK_AUTO_AUDIT_PERCENTAGE,
    CONSENSUS_AUTO_AUDIT_NUMBER_OF_LABELS,
    CONSENSUS_AUTO_AUDIT_PERCENTAGE,
    QualityMode,
)

PositiveInt = Annotated[int, Field(gt=0)]


class _CoreProjectInput(BaseModel):
    name: str
    description: Optional[str] = None
    media_type: MediaType
    auto_audit_percentage: Optional[float] = None
    auto_audit_number_of_labels: Optional[int] = None
    quality_modes: Optional[Set[QualityMode]] = Field(
        default={QualityMode.Benchmark, QualityMode.Consensus}, exclude=True
    )
    is_benchmark_enabled: Optional[bool] = None
    is_consensus_enabled: Optional[bool] = None
    dataset_name_or_id: Optional[str] = None
    append_to_existing_dataset: Optional[bool] = None
    data_row_count: Optional[PositiveInt] = None
    editor_task_type: Optional[EditorTaskType] = None

    model_config = ConfigDict(extra="forbid", use_enum_values=True)

    @model_validator(mode="after")
    def validate_fields(self):
        if (
            self.auto_audit_percentage is not None
            or self.auto_audit_number_of_labels is not None
        ):
            raise ValueError(
                "quality_modes must be set instead of auto_audit_percentage or auto_audit_number_of_labels."
            )

        if not self.name.strip():
            raise ValueError("project name must be a valid string.")

        if self.quality_modes == {
            QualityMode.Benchmark,
            QualityMode.Consensus,
        }:
            self._set_quality_mode_attributes(
                CONSENSUS_AUTO_AUDIT_NUMBER_OF_LABELS,
                CONSENSUS_AUTO_AUDIT_PERCENTAGE,
                is_benchmark_enabled=True,
                is_consensus_enabled=True,
            )
        elif self.quality_modes == {QualityMode.Benchmark}:
            self._set_quality_mode_attributes(
                BENCHMARK_AUTO_AUDIT_NUMBER_OF_LABELS,
                BENCHMARK_AUTO_AUDIT_PERCENTAGE,
                is_benchmark_enabled=True,
            )
        elif self.quality_modes == {QualityMode.Consensus}:
            self._set_quality_mode_attributes(
                number_of_labels=CONSENSUS_AUTO_AUDIT_NUMBER_OF_LABELS,
                percentage=CONSENSUS_AUTO_AUDIT_PERCENTAGE,
                is_consensus_enabled=True,
            )

        if self.data_row_count is not None and self.data_row_count < 0:
            raise ValueError("data_row_count must be a positive integer.")

        return self

    def _set_quality_mode_attributes(
        self,
        number_of_labels,
        percentage,
        is_benchmark_enabled=False,
        is_consensus_enabled=False,
    ):
        self.auto_audit_number_of_labels = number_of_labels
        self.auto_audit_percentage = percentage
        self.is_benchmark_enabled = is_benchmark_enabled
        self.is_consensus_enabled = is_consensus_enabled



def test_enable_model_assisted_labeling(project):
    response = project.enable_model_assisted_labeling()
    assert response is True

    response = project.enable_model_assisted_labeling(True)
    assert response is True

    response = project.enable_model_assisted_labeling(False)
    assert response is False



import json
import logging
from datetime import datetime, timezone
from functools import wraps

from lbox.exceptions import (
    InvalidAttributeError,
    InvalidQueryError,
    OperationNotSupportedException,
)

from labelbox import utils
from labelbox.orm import query
from labelbox.orm.model import Entity, Field, Relationship
from labelbox.pagination import PaginatedCollection

logger = logging.getLogger(__name__)


class DbObject(Entity):
    """A client-side representation of a database object (row). Intended as
    base class for classes representing concrete database types (for example
    a Project). Exposes support functionalities so that the concrete subclass
    definition be as simple and DRY as possible. It should come down to just
    listing Fields of that particular database type. For example:

        >>> class Project(DbObject):
        >>>     name = Field.String("name")
        >>>     labels = Relationship.ToMany("Label", True)

    This defines a `Project` class that has class attributes which are
    `Field`s and `Relationship`s. An instance of `Project` represents
    a database record. It has the same attributes as the `Project` class,
    but they are now attribute values of that record:

        >>> project = client.create_project(name="MyProject")
        >>> project.name
        "MyProject"
    """

    def __init__(self, client, field_values):
        """Constructor of a database object. Generally it should only be used
        by library internals and not by the end user.

        Args:
            client (labelbox.Client): the client used for fetching data from DB.
            field_values (dict): Data obtained from the DB. Maps database object
                fields (their graphql_name version) to values.
        """
        self.client = client
        self._set_field_values(field_values)
        for relationship in self.relationships():
            value = field_values.get(utils.camel_case(relationship.name))
            if relationship.cache and value is None:
                raise KeyError(
                    f"Expected field  values for {relationship.name}"
                )
            setattr(
                self,
                relationship.name,
                RelationshipManager(self, relationship, value),
            )

    def _set_field_values(self, field_values):
        """Sets field values on this object. Ensures proper value conversions.
        Args:
            field_values (dict): Maps field names (GraphQL variant, snakeCase)
                to values. *Must* contain all field values for this object's
                DB type.
        """
        for field in self.fields():
            value = field_values[field.graphql_name]
            if field.field_type == Field.Type.DateTime and value is not None:
                try:
                    value = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
                    value = value.replace(tzinfo=timezone.utc)
                except ValueError:
                    logger.warning(
                        "Failed to convert value '%s' to datetime for "
                        "field %s",
                        value,
                        field,
                    )
            elif isinstance(field.field_type, Field.EnumType):
                value = field.field_type.enum_cls(value)
            elif isinstance(field.field_type, Field.ListType):
                if field.field_type.list_cls.__name__ == "DataRowMetadataField":
                    mdo = self.client.get_data_row_metadata_ontology()
                    try:
                        value = mdo.parse_metadata_fields(value)
                    except ValueError:
                        logger.warning(
                            "Failed to convert value '%s' to metadata for field %s",
                            value,
                            field,
                        )
            setattr(self, field.name, value)

    def __repr__(self):
        type_name = self.type_name()
        if "uid" in self.__dict__:
            return "<%s ID: %s>" % (type_name, self.uid)
        else:
            return "<%s>" % type_name

    def __str__(self):
        attribute_values = {
            field.name: getattr(self, field.name) for field in self.fields()
        }
        return "<%s %s>" % (
            self.type_name().split(".")[-1],
            json.dumps(attribute_values, indent=4, default=str),
        )

    def __eq__(self, other):
        return (
            isinstance(other, DbObject)
            and self.type_name() == other.type_name()
            and self.uid == other.uid
        )

    def __hash__(self):
        return 7541 * hash(self.type_name()) + hash(self.uid)


class RelationshipManager:
    """Manages relationships (object fetching and updates) for a `DbObject`
    instance. There is one RelationshipManager for each relationship in
    each `DbObject` instance.
    """

    def __init__(self, source, relationship, value=None):
        """Args:
        source (DbObject subclass instance): The object that's the source
            of the relationship.
        relationship (labelbox.schema.Relationship): The relationship
            schema descriptor object.
        """
        self.source = source
        self.relationship = relationship
        self.supports_filtering = True
        self.supports_sorting = True
        self.filter_on_id = True
        self.value = value
        self.config = relationship.config

    def __call__(self, *args, **kwargs):
        """Forwards the call to either `_to_many` or `_to_one` methods,
        depending on relationship type."""

        if self.relationship.deprecation_warning:
            logger.warning(self.relationship.deprecation_warning)

        if self.relationship.relationship_type == Relationship.Type.ToMany:
            return self._to_many(*args, **kwargs)
        else:
            return self._to_one(*args, **kwargs)

    def _to_many(self, where=None, order_by=None):
        """Returns an iterable over the destination relationship objects.
        Args:
            where (None, Comparison or LogicalExpression): Filtering clause.
            order_by (None or (Field, Field.Order)): Ordering clause.
        Return:
            iterable over destination DbObject instances.
        """
        rel = self.relationship
        if where is not None and not self.supports_filtering:
            raise InvalidQueryError(
                "Relationship %s.%s doesn't support filtering"
                % (self.source.type_name(), rel.name)
            )
        if order_by is not None and not self.supports_sorting:
            raise InvalidQueryError(
                "Relationship %s.%s doesn't support sorting"
                % (self.source.type_name(), rel.name)
            )

        if rel.filter_deleted:
            not_deleted = rel.destination_type.deleted == False  # noqa: E712 <Gabefire> Needed for bit operator to combine comparisons
            where = not_deleted if where is None else where & not_deleted

        query_string, params = query.relationship(
            self.source if self.filter_on_id else type(self.source),
            rel,
            where,
            order_by,
        )
        return PaginatedCollection(
            self.source.client,
            query_string,
            params,
            [utils.camel_case(self.source.type_name()), rel.graphql_name],
            rel.destination_type,
        )

    def _to_one(self):
        """Returns the relationship destination object."""
        rel = self.relationship

        if self.value:
            return rel.destination_type(self.source.client, self.value)

        query_string, params = query.relationship(self.source, rel, None, None)
        result = self.source.client.execute(query_string, params)
        result = result and result.get(
            utils.camel_case(type(self.source).type_name())
        )
        result = result and result.get(rel.graphql_name)
        if result is None:
            return None

        return rel.destination_type(self.source.client, result)

    def connect(self, other):
        """Connects source object of this manager to the `other` object."""
        query_string, params = query.update_relationship(
            self.source, other, self.relationship, "connect"
        )
        self.source.client.execute(query_string, params)

    def disconnect(self, other):
        """Disconnects source object of this manager from the `other` object."""
        if not self.config.disconnect_supported:
            raise OperationNotSupportedException(
                "Disconnect is not supported for this relationship"
            )

        query_string, params = query.update_relationship(
            self.source, other, self.relationship, "disconnect"
        )
        self.source.client.execute(query_string, params)


class Updateable:
    def update(self, **kwargs):
        """Updates this DB object with new values. Values should be
        passed as key-value arguments with field names as keys:
        >>> db_object.update(name="New name", title="A title")

        Kwargs:
            Key-value arguments defining which fields should be updated
            for which values. Keys must be field names in this DB object's
            type.
        Raise:
            InvalidAttributeError: if there exists a key in `kwargs`
                that's not a field in this object type.
        """
        values = {self.field(name): value for name, value in kwargs.items()}
        invalid_fields = set(values) - set(self.fields())
        if invalid_fields:
            raise InvalidAttributeError(type(self), invalid_fields)

        query_string, params = query.update_fields(self, values)
        res = self.client.execute(query_string, params)
        res = res["update%s" % utils.title_case(self.type_name())]
        self._set_field_values(res)


class Deletable:
    """Implements deletion for objects that have a `deleted` attribute."""

    def delete(self):
        """Deletes this DB object from the DB (server side). After
        a call to this you should not use this DB object anymore.
        """
        query_string, params = query.delete(self)
        self.client.execute(query_string, params)


class BulkDeletable:
    """Implements deletion for objects that have a custom, bulk deletion
    mutation (accepts a list of IDs of objects to be deleted).

    A subclass must override the `bulk_delete` static method so it
    accepts only the `objects` argument and then invoke BulkDeletable.bulk_delete
    with the appropriate `use_where_clause` argument for that particular
    type.
    """

    @staticmethod
    def _bulk_delete(objects, use_where_clause):
        """
        Args:
            objects (list): Objects to delete. All objects must be of the same
                DbObject subtype.
            use_where_clause (bool): If the GraphQL query object IDs should be
                passed under `where` or directly. Necessary because the bulk
                deletion mutation is implemented differently for different
                object types (DataRow.bulkDelete vs Label.bulkDelete).
        """
        types = {type(o) for o in objects}
        if len(types) != 1:
            raise InvalidQueryError(
                "Can't bulk-delete objects of different types: %r" % types
            )

        query_str, params = query.bulk_delete(objects, use_where_clause)
        objects[0].client.execute(query_str, params)

    def delete(self):
        """Deletes this DB object from the DB (server side). After
        a call to this you should not use this DB object anymore.
        """
        type(self).bulk_delete([self])


def experimental(fn):
    """Decorator used to mark functions that are experimental. This means that
    the interface could change. This decorator will check if the client has
    experimental features enabled. If not, it will raise a runtime error.
    """
    is_static = isinstance(fn, staticmethod)

    @wraps(fn)
    def wrapper(*args, **kwargs):
        client = None
        wrapped_fn = None
        if is_static:
            # assumes that the first argument is the client, needs modification if that changes
            if len(args) >= 1:
                client = args[0]
            elif "client" in kwargs:
                client = kwargs["client"]
            else:
                raise ValueError(
                    f"Static method {fn.__name__} must have a client passed in as the first "
                    f"argument or as a keyword argument."
                )
            wrapped_fn = fn.__func__
        else:
            client = args[0].client
            wrapped_fn = fn

        if not client.enable_experimental:
            raise RuntimeError(
                f"This function {fn.__name__} relies on a experimental feature in the api. "
                f"This means that the interface could change. "
                f"Set `enable_experimental=True` in the client to enable use of "
                f"experimental functions."
            )
        return wrapped_fn(*args, **kwargs)

    return wrapper



from labelbox.schema.tool_building.base_step_reasoning_tool import _Variant


def test_variant():
    variant = _Variant(
        id=0, name="Correct", _available_actions={"regenerateSteps"}
    )
    variant.set_actions(["regenerateSteps"])
    assert variant.asdict() == {
        "id": 0,
        "name": "Correct",
        "actions": ["regenerateSteps"],
    }

    assert variant._available_actions == {"regenerateSteps"}
    variant.reset_actions()
    assert variant.asdict() == {
        "id": 0,
        "name": "Correct",
        "actions": [],
    }


def test_variant_actions():
    variant = _Variant(
        id=0, name="Correct", _available_actions={"regenerateSteps"}
    )
    variant.set_actions(["regenerateSteps"])
    assert variant.actions == ["regenerateSteps"]
    variant.set_actions(["invalidAction"])
    assert variant.actions == []



import pytest
from lbox.exceptions import error_message_for_unparsed_graphql_error


@pytest.mark.parametrize(
    "exception_message, expected_result",
    [
        (
            "Unparsed errors on query execution: [{'message': 'Cannot create model config for project because model setup is complete'}]",
            "Cannot create model config for project because model setup is complete",
        ),
        ("blah blah blah", "Unknown error"),
    ],
)
def test_client_unparsed_exception_messages(exception_message, expected_result):
    assert (
        error_message_for_unparsed_graphql_error(exception_message)
        == expected_result
    )



from typing import Dict, Any

from pydantic import BaseModel, model_validator


class TextEntity(BaseModel):
    """Represents a text entity"""

    start: int
    end: int
    extra: Dict[str, Any] = {}

    @model_validator(mode="after")
    def validate_start_end(self, values):
        if hasattr(self, "start") and hasattr(self, "end"):
            if isinstance(self.start, int) and self.start > self.end:
                raise ValueError(
                    "Location end must be greater or equal to start"
                )
        return self



from dataclasses import dataclass
from typing import Dict, Optional, TYPE_CHECKING

from labelbox.orm.model import Field
from labelbox.orm.db_object import DbObject

if TYPE_CHECKING:
    from labelbox import Client, Project

_ROLES: Optional[Dict[str, "Role"]] = None


def get_roles(client: "Client") -> Dict[str, "Role"]:
    global _ROLES
    if _ROLES is None:
        query_str = """query GetAvailableUserRolesPyApi { roles { id name } }"""
        res = client.execute(query_str)
        _ROLES = {}
        for role in res["roles"]:
            role["name"] = format_role(role["name"])
            _ROLES[role["name"]] = Role(client, role)
    return _ROLES


def format_role(name: str):
    return name.upper().replace(" ", "_")


class Role(DbObject):
    name = Field.String("name")


class OrgRole(Role): ...


class UserRole(Role): ...


@dataclass
class ProjectRole:
    project: "Project"
    role: Role



from typing import List, Optional

from lbox.exceptions import (
    ConfidenceNotSupportedException,
    CustomMetricsNotSupportedException,
)
from pydantic import BaseModel, field_validator


class ConfidenceMixin(BaseModel):
    confidence: Optional[float] = None

    @field_validator("confidence")
    def confidence_valid_float(cls, value):
        if value is None:
            return value
        if not isinstance(value, (int, float)) or not 0 <= value <= 1:
            raise ValueError("must be a number within [0,1] range")
        return value


class ConfidenceNotSupportedMixin:
    def __new__(cls, *args, **kwargs):
        if "confidence" in kwargs:
            raise ConfidenceNotSupportedException(
                "Confidence is not supported for this annotation type yet"
            )
        return super().__new__(cls)


class CustomMetric(BaseModel):
    name: str
    value: float

    @field_validator("name")
    def confidence_valid_float(cls, value):
        if not isinstance(value, str):
            raise ValueError("Name must be a string")
        return value

    @field_validator("value")
    def value_valid_float(cls, value):
        if not isinstance(value, (int, float)):
            raise ValueError("Value must be a number")
        return value


class CustomMetricsMixin(BaseModel):
    custom_metrics: Optional[List[CustomMetric]] = None


class CustomMetricsNotSupportedMixin:
    def __new__(cls, *args, **kwargs):
        if "custom_metrics" in kwargs:
            raise CustomMetricsNotSupportedException(
                "Custom metrics is not supported for this annotation type yet"
            )
        return super().__new__(cls)



try:
    from labelbox.data.annotation_types import *
except ImportError:
    raise ImportError(
        "There are missing dependencies for `labelbox.types`, use `pip install labelbox[data] --upgrade` to install missing dependencies."
    )



import json
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import TYPE_CHECKING, Generator, Iterable, List

from lbox.exceptions import (
    InvalidAttributeError,
    InvalidQueryError,
)  # type: ignore

from labelbox.orm.model import Entity, Field
from labelbox.schema.embedding import EmbeddingVector
from labelbox.schema.internal.data_row_upsert_item import (
    DataRowItemBase,
)
from labelbox.schema.internal.datarow_upload_constants import (
    FILE_UPLOAD_THREAD_COUNT,
)

if TYPE_CHECKING:
    from labelbox import Client


class DescriptorFileCreator:
    """
    This class is used to upload a list of dict and return a url.
    It will create multiple files if the size of upload the max_chunk_size in bytes,
        upload the files to gcs in parallel, and return a list of urls

    Args:
        client (Client): The client object
        max_chunk_size_bytes (int): The maximum size of the file in bytes
    """

    def __init__(self, client: "Client"):
        self.client = client

    def create(self, items, max_chunk_size_bytes=None) -> List[str]:
        is_upsert = True  # This class will only support upsert use cases
        items = self._prepare_items_for_upload(items, is_upsert=is_upsert)
        json_chunks = self._chunk_down_by_bytes(items, max_chunk_size_bytes)
        with ThreadPoolExecutor(FILE_UPLOAD_THREAD_COUNT) as executor:
            futures = [
                executor.submit(
                    self.client.upload_data,
                    chunk,
                    "application/json",
                    "json_import.json",
                )
                for chunk in json_chunks
            ]
            return [future.result() for future in as_completed(futures)]

    def create_one(self, items) -> List[str]:
        items = self._prepare_items_for_upload(
            items,
        )
        # Prepare and upload the descriptor file
        data = json.dumps(items)
        return self.client.upload_data(
            data, content_type="application/json", filename="json_import.json"
        )

    def _prepare_items_for_upload(self, items, is_upsert=False):
        """
        This function is used to prepare the input file. The user defined input is validated, processed, and json stringified.
        Finally the json data is uploaded to gcs and a uri is returned. This uri can be passed as a parameter to a mutation that uploads data rows

        Each element in `items` can be either a `str` or a `dict`. If
        it is a `str`, then it is interpreted as a local file path. The file
        is uploaded to Labelbox and a DataRow referencing it is created.

        If an item is a `dict`, then it could support one of the two following structures
            1. For static imagery, video, and text it should map `DataRow` field names to values.
                At the minimum an `items` passed as a `dict` must contain a `row_data` key and value.
                If the value for row_data is a local file path and the path exists,
                then the local file will be uploaded to labelbox.

            2. For tiled imagery the dict must match the import structure specified in the link below
                https://docs.labelbox.com/data-model/en/index-en#tiled-imagery-import

        >>> dataset.create_data_rows([
        >>>     {DataRow.row_data:"http://my_site.com/photos/img_01.jpg"},
        >>>     {DataRow.row_data:"/path/to/file1.jpg"},
        >>>     "path/to/file2.jpg",
        >>>     {DataRow.row_data: {"tileLayerUrl" : "http://", ...}}
        >>>     {DataRow.row_data: {"type" : ..., 'version' : ..., 'messages' : [...]}}
        >>>     ])

        Args:
            items (iterable of (dict or str)): See above for details.

        Returns:
            uri (string): A reference to the uploaded json data.

        Raises:
            InvalidQueryError: If the `items` parameter does not conform to
                the specification above or if the server did not accept the
                DataRow creation request (unknown reason).
            InvalidAttributeError: If there are fields in `items` not valid for
                a DataRow.
            ValueError: When the upload parameters are invalid
        """
        file_upload_thread_count = FILE_UPLOAD_THREAD_COUNT
        DataRow = Entity.DataRow
        AssetAttachment = Entity.AssetAttachment

        def upload_if_necessary(item):
            if is_upsert and "row_data" not in item:
                # When upserting, row_data is not required
                return item
            row_data = item["row_data"]
            if isinstance(row_data, str) and os.path.exists(row_data):
                item_url = self.client.upload_file(row_data)
                item["row_data"] = item_url
                if "external_id" not in item:
                    # Default `external_id` to local file name
                    item["external_id"] = row_data
            return item

        def validate_attachments(item):
            attachments = item.get("attachments")
            if attachments:
                if isinstance(attachments, list):
                    for attachment in attachments:
                        AssetAttachment.validate_attachment_json(attachment)
                else:
                    raise ValueError(
                        f"Attachments must be a list. Found {type(attachments)}"
                    )
            return attachments

        def validate_embeddings(item):
            embeddings = item.get("embeddings")
            if embeddings:
                item["embeddings"] = [
                    EmbeddingVector(**e).to_gql() for e in embeddings
                ]

        def validate_conversational_data(conversational_data: list) -> None:
            """
            Checks each conversational message for keys expected as per https://docs.labelbox.com/reference/text-conversational#sample-conversational-json

            Args:
                conversational_data (list): list of dictionaries.
            """

            def check_message_keys(message):
                accepted_message_keys = set(
                    [
                        "messageId",
                        "timestampUsec",
                        "content",
                        "user",
                        "align",
                        "canLabel",
                    ]
                )
                for key in message.keys():
                    if key not in accepted_message_keys:
                        raise KeyError(
                            f"Invalid {key} key found! Accepted keys in messages list is {accepted_message_keys}"
                        )

            if conversational_data and not isinstance(
                conversational_data, list
            ):
                raise ValueError(
                    f"conversationalData must be a list. Found {type(conversational_data)}"
                )

            [check_message_keys(message) for message in conversational_data]

        def parse_metadata_fields(item):
            metadata_fields = item.get("metadata_fields")
            if metadata_fields:
                mdo = self.client.get_data_row_metadata_ontology()
                item["metadata_fields"] = mdo.parse_upsert_metadata(
                    metadata_fields
                )

        def format_row(item):
            # Formats user input into a consistent dict structure
            if isinstance(item, dict):
                # Convert fields to strings
                item = {
                    key.name if isinstance(key, Field) else key: value
                    for key, value in item.items()
                }
            elif isinstance(item, str):
                # The main advantage of using a string over a dict is that the user is specifying
                # that the file should exist locally.
                # That info is lost after this section so we should check for it here.
                if not os.path.exists(item):
                    raise ValueError(f"Filepath {item} does not exist.")
                item = {"row_data": item, "external_id": item}
            return item

        def validate_keys(item):
            if not is_upsert and "row_data" not in item:
                raise InvalidQueryError(
                    "`row_data` missing when creating DataRow."
                )

            if isinstance(item.get("row_data"), str) and item.get(
                "row_data"
            ).startswith("s3:/"):
                raise InvalidQueryError(
                    "row_data: s3 assets must start with 'https'."
                )
            allowed_extra_fields = {
                "attachments",
                "media_type",
                "dataset_id",
                "embeddings",
            }
            invalid_keys = (
                set(item)
                - {f.name for f in DataRow.fields()}
                - allowed_extra_fields
            )
            if invalid_keys:
                raise InvalidAttributeError(DataRow, invalid_keys)
            return item

        def format_legacy_conversational_data(item):
            messages = item.pop("conversationalData")
            version = item.pop("version", 1)
            type = item.pop("type", "application/vnd.labelbox.conversational")
            if "externalId" in item:
                external_id = item.pop("externalId")
                item["external_id"] = external_id
            if "globalKey" in item:
                global_key = item.pop("globalKey")
                item["globalKey"] = global_key
            validate_conversational_data(messages)
            one_conversation = {
                "type": type,
                "version": version,
                "messages": messages,
            }
            item["row_data"] = one_conversation
            return item

        def convert_item(data_row_item):
            if isinstance(data_row_item, DataRowItemBase):
                item = data_row_item.payload
            else:
                item = data_row_item

            if "tileLayerUrl" in item:
                validate_attachments(item)
                return item

            if "conversationalData" in item:
                format_legacy_conversational_data(item)

            # Convert all payload variations into the same dict format
            item = format_row(item)
            # Make sure required keys exist (and there are no extra keys)
            validate_keys(item)
            # Make sure attachments are valid
            validate_attachments(item)
            # Make sure embeddings are valid
            validate_embeddings(item)
            # Parse metadata fields if they exist
            parse_metadata_fields(item)
            # Upload any local file paths
            item = upload_if_necessary(item)

            if isinstance(data_row_item, DataRowItemBase):
                return {"id": data_row_item.id, "payload": item}
            else:
                return item

        if not isinstance(items, Iterable):
            raise ValueError(
                f"Must pass an iterable to create_data_rows. Found {type(items)}"
            )

        with ThreadPoolExecutor(file_upload_thread_count) as executor:
            futures = [executor.submit(convert_item, item) for item in items]
            items = [future.result() for future in as_completed(futures)]

        return items

    def _chunk_down_by_bytes(
        self, items: List[dict], max_chunk_size: int
    ) -> Generator[str, None, None]:
        """
        Recursively chunks down a list of items into smaller lists until each list is less than or equal to max_chunk_size bytes
        NOTE: if one data row is larger than max_chunk_size, it will be returned as one chunk

        Returns:
            Generator[str, None, None]: A generator that yields a json string
        """
        if not items:
            return
        data = json.dumps(items)
        chunk_size = len(data.encode("utf-8"))
        if chunk_size <= max_chunk_size:
            yield data
            return

        if len(items) == 1:
            yield data
            return

        half = len(items) // 2
        yield from self._chunk_down_by_bytes(items[:half], max_chunk_size)
        yield from self._chunk_down_by_bytes(items[half:], max_chunk_size)



from itertools import chain
from typing import Any, Dict

from lbox.exceptions import (
    InvalidAttributeError,
    InvalidQueryError,
    MalformedQueryException,
)

from labelbox import utils
from labelbox.orm.comparison import Comparison, LogicalExpression
from labelbox.orm.model import Entity, Field, Relationship

""" Common query creation functionality. """


def format_param_declaration(params):
    """Formats the parameters dictionary into a declaration of GraphQL
    query parameters.

    Args:
        params (dict): keys are query param names and values are
            (value, (field|relationship)) tuples.
    Return:
        str, the declaration of query parameters.
    """
    if not params:
        return ""

    def attr_type(attr):
        if isinstance(attr, Field):
            return attr.field_type.name
        else:
            return Field.Type.ID.name

    return (
        "("
        + ", ".join(
            "$%s: %s!" % (param, attr_type(attr))
            for param, (_, attr) in params.items()
        )
        + ")"
    )


def results_query_part(entity):
    """Generates the results part of the query. The results contain
    all the entity's fields as well as prefetched relationships.

    Note that this is a recursive function. If there is a cycle in the
    prefetched relationship graph, this function will recurse infinitely.

    Args:
        entity (type): The entity which needs fetching.
    """
    # Query for fields
    fields = [
        field.result_subquery
        if field.result_subquery is not None
        else field.graphql_name
        for field in entity.fields()
    ]

    # Query for cached relationships
    fields.extend(
        [
            Query(rel.graphql_name, rel.destination_type).format()[0]
            for rel in entity.relationships()
            if rel.cache
        ]
    )
    return " ".join(fields)


class Query:
    """A data structure used during the construction of a query. Supports
    subquery (also Query object) nesting for relationship."""

    def __init__(
        self, what, subquery, where=None, paginate=False, order_by=None
    ):
        """Initializer.
        Args:
            what (str): What is being queried. Typically an object type in
                singular or plural (i.e. "project" or "projects").
            subquery (Query or type): Either a Query object that is formatted
                recursively or a Entity subtype in which case the standard
                results (see `results_query_part`) are retrieved.
            where (None, Comparison or LogicalExpression): the filtering clause.
            paginate (bool): If the "%skip %first" pagination substring should
                be added to the query. Used for collection pagination in combination
                with PaginatedCollection.
            order_by (tuple): A tuple consisting of (Field, Field.Order) indicating
                how the query should sort the collection.
        """
        self.what = what
        self.subquery = subquery
        self.paginate = paginate
        self.where = where
        self.order_by = order_by

    def format_subquery(self):
        """Formats the subquery (a Query or Entity subtype)."""
        if isinstance(self.subquery, Query):
            return self.subquery.format()
        elif issubclass(self.subquery, Entity):
            return results_query_part(self.subquery), {}
        else:
            raise MalformedQueryException()

    def format_clauses(self, params):
        """Formats the where, order_by and pagination clauses.
        Args:
            params (dict): The current parameter dictionary.
        """

        def format_where(node):
            """Helper that resursively constructs a where clause from a
            LogicalExpression tree (leaf nodes are Comparisons)."""
            COMPARISON_TO_SUFFIX = {
                Comparison.Op.EQ: "",
                Comparison.Op.NE: "_not",
                Comparison.Op.LT: "_lt",
                Comparison.Op.GT: "_gt",
                Comparison.Op.LE: "_lte",
                Comparison.Op.GE: "_gte",
            }
            assert isinstance(node, (Comparison, LogicalExpression))
            if isinstance(node, Comparison):
                param_name = "param_%d" % len(params)
                params[param_name] = (node.value, node.field)
                return "{%s%s: $%s}" % (
                    node.field.graphql_name,
                    COMPARISON_TO_SUFFIX[node.op],
                    param_name,
                )
            if node.op == LogicalExpression.Op.NOT:
                return "{NOT: [%s]}" % format_where(node.first)

            return "{%s: [%s, %s]}" % (
                node.op.name.upper(),
                format_where(node.first),
                format_where(node.second),
            )

        paginate = "skip: %d first: %d" if self.paginate else ""

        where = "where: %s" % format_where(self.where) if self.where else ""

        if self.order_by:
            order_by = "orderBy: %s_%s" % (
                self.order_by[0].graphql_name,
                self.order_by[1].name.upper(),
            )
        else:
            order_by = ""

        clauses = " ".join(filter(None, (where, paginate, order_by)))
        return "(" + clauses + ")" if clauses else ""

    def format(self):
        """Formats the full query but without "query" prefix, query name
        and parameter declaration.
        Return:
            (str, dict) tuple. str is the query and dict maps parameter
            names to (value, field) tuples.
        """
        subquery, params = self.format_subquery()
        clauses = self.format_clauses(params)
        query = "%s%s{%s}" % (self.what, clauses, subquery)
        return query, params

    def format_top(self, name):
        """Formats the full query including "query" prefix, query name
        and parameter declaration. The result of this function can be
        sent to the Client object for execution.

        Args:
            name (str): Query name, without the "PyApi" suffix, it's appended
                automatically by this method.
        Return:
            (str, dict) tuple. str is the full query and dict maps parameter
                names to parameter values.
        """
        query, params = self.format()
        param_declaration = format_param_declaration(params)
        query = "query %sPyApi%s{%s}" % (name, param_declaration, query)
        return query, {param: value for param, (value, _) in params.items()}


def get_single(entity, uid):
    """Constructs the query and params dict for obtaining a single object. Either
    on ID, or without params.
    Args:
        entity (type): An Entity subtype being obtained.
        uid (str): The ID of the sought object. It can be None, which is legal for
            DB types that have a default object being returned (User and
            Organization).
    """
    type_name = entity.type_name()
    where = entity.uid == uid if uid else None
    return Query(utils.camel_case(type_name), entity, where).format_top(
        "Get" + type_name
    )


def logical_ops(where):
    """Returns a generator that yields all the logical operator
    type objects (`LogicalExpression.Op` instances) from a where
    clause.

    Args:
        where (LogicalExpression, Comparison or None): The where
            clause used for filtering in a query.
    Return:
        See above.
    """
    if isinstance(where, LogicalExpression):
        yield where.op
        for f in chain(logical_ops(where.first), logical_ops(where.second)):
            yield f


def check_where_clause(entity, where):
    """Checks the `where` clause of a query. A `where` clause is legal
    if it only refers to fields found in the entity it's defined for.
    Since only AND logical operations are supported server-side at the
    moment, logical OR and NOT are illegal.

    Args:
        entity (type): An Entity subclass type.
        where (LogicalExpression or Comparison): The `where` clause of
            query.
    Return:
        bool indicating if `where` is legal for `entity`.
    """

    def fields(where):
        """Yields all the fields in a `where` clause."""
        if isinstance(where, LogicalExpression):
            for f in chain(fields(where.first), fields(where.second)):
                yield f
        elif isinstance(where, Comparison):
            yield where.field

    # The `deleted` field is a special case, ignore it.
    where_fields = [f for f in fields(where) if f != Entity.deleted]
    invalid_fields = set(where_fields) - set(entity.fields())
    if invalid_fields:
        raise InvalidAttributeError(entity, invalid_fields)

    if len(set(where_fields)) != len(where_fields):
        raise InvalidQueryError(
            "Where clause contains multiple comparisons for "
            "the same field: %r." % where
        )

    if set(logical_ops(where)) not in (set(), {LogicalExpression.Op.AND}):
        raise InvalidQueryError(
            "Currently only AND logical ops are allowed in "
            "the where clause of a query."
        )


def check_order_by_clause(entity, order_by):
    """Checks that the `order_by` clause field is a part of `entity`.

    Args:
        entity (type): An Entity subclass type.
        order_by ((field, ordering)): The ordering tuple consisting of
            a field and sort ordering (ascending or descending).
    Return:
        bool indicating if `order_by` is legal for `entity`.
    """
    if order_by is not None:
        field, _ = order_by
        if field not in entity.fields():
            raise InvalidAttributeError(entity, field)


def get_all(entity, where):
    """Constructs a query that fetches all items of the given type. The
    resulting query is intended to be used for pagination, it contains
    two python-string int-placeholders (%d) for 'skip' and 'first'
    pagination parameters.

    Args:
        entity (type): The object type being queried.
        where (Comparison, LogicalExpression or None): The `where` clause
            for filtering.
    Return:
        (str, dict) tuple that is the query string and parameters.
    """
    check_where_clause(entity, where)
    type_name = entity.type_name()
    query = Query(utils.camel_case(type_name) + "s", entity, where, True)
    return query.format_top("Get" + type_name + "s")


def relationship(source, relationship, where, order_by):
    """Constructs a query that fetches all items from a -to-many
    relationship. To be used like:
        >>> project = ...
        >>> query_str, params = relationship(Project, "datasets", Dataset)
        >>> datasets = PaginatedCollection(
            client, query_str, params, ["project", "datasets"],
            Dataset)

    The resulting query is intended to be used for pagination, it contains
    two python-string int-placeholders (%d) for 'skip' and 'first'
    pagination parameters.

    Args:
        source (DbObject or type): If a `DbObject` then the source of the
            relationship (the query originates from that particular object).
            If `type`, then the source of the relationship is implicit, even
            without the ID. Used for expanding from Organization.
        relationship (Relationship): The relationship.
        where (Comparison, LogicalExpression or None): The `where` clause
            for filtering.
        order_by (None or (Field, Field.Order): The `order_by` clause for
            sorting results.
    Return:
        (str, dict) tuple that is the query string and parameters.
    """
    check_where_clause(relationship.destination_type, where)
    check_order_by_clause(relationship.destination_type, order_by)
    to_many = relationship.relationship_type == Relationship.Type.ToMany
    subquery = Query(
        relationship.graphql_name,
        relationship.destination_type,
        where,
        to_many,
        order_by,
    )
    query_where = (
        type(source).uid == source.uid if isinstance(source, Entity) else None
    )
    query = Query(utils.camel_case(source.type_name()), subquery, query_where)
    return query.format_top(
        "Get" + source.type_name() + utils.title_case(relationship.graphql_name)
    )


def create(entity, data):
    """Generates a query and parameters for creating a new DB object.

    Args:
        entity (type): An Entity subtype indicating which kind of
            DB object needs to be created.
        data (dict): A dict that maps Fields and Relationships to values, new
            object data.
    Return:
        (query_string, parameters)
    """
    type_name = entity.type_name()

    def format_param_value(attribute, param):
        if isinstance(attribute, Field):
            return "%s: $%s" % (attribute.graphql_name, param)
        else:
            return "%s: {connect: {id: $%s}}" % (
                utils.camel_case(attribute.graphql_name),
                param,
            )

    # Convert data to params
    params = {
        field.graphql_name: (value, field) for field, value in data.items()
    }

    query_str = """mutation Create%sPyApi%s{create%s(data: {%s}) {%s}} """ % (
        type_name,
        format_param_declaration(params),
        type_name,
        " ".join(
            format_param_value(attribute, param)
            for param, (_, attribute) in params.items()
        ),
        results_query_part(entity),
    )

    return query_str, {name: value for name, (value, _) in params.items()}


def update_relationship(a, b, relationship, update):
    """Updates the relationship in DB object `a` to connect or disconnect
    DB object `b`.

    Args:
        a (DbObject): The object being updated.
        b (DbObject): Object on the other side of the relationship.
        relationship (Relationship): The relationship from `a` to `b`.
        update (str): The type of update. Must be either `connect` or
            `disconnect`.
    Return:
        (query_string, query_parameters)
    """
    to_one_disconnect = (
        update == "disconnect"
        and relationship.relationship_type == Relationship.Type.ToOne
    )

    a_uid_param = utils.camel_case(type(a).type_name()) + "Id"

    if not to_one_disconnect:
        b_uid_param = utils.camel_case(type(b).type_name()) + "Id"
        param_declr = "($%s: ID!, $%s: ID!)" % (a_uid_param, b_uid_param)
        b_query = "{id: $%s}" % b_uid_param
    else:
        param_declr = "($%s: ID!)" % a_uid_param
        b_query = "true"

    query_str = """mutation %s%sAnd%sPyApi%s{update%s(
        where: {id: $%s} data: {%s: {%s: %s}}) {id}} """ % (
        utils.title_case(update),
        type(a).type_name(),
        type(b).type_name(),
        param_declr,
        utils.title_case(type(a).type_name()),
        a_uid_param,
        relationship.graphql_name,
        update,
        b_query,
    )

    if to_one_disconnect:
        params = {a_uid_param: a.uid}
    else:
        params = {a_uid_param: a.uid, b_uid_param: b.uid}

    return query_str, params


def update_fields(db_object, values):
    """Creates a query that updates `db_object` fields with the
    given values.

    Args:
        db_object (DbObject): The DB object being updated.
        values (dict): Maps Fields to new values. All Fields
            must be legit fields in `db_object`.
    Return:
        (query_string, query_parameters)
    """
    type_name = db_object.type_name()
    id_param = "%sId" % type_name
    values_str = " ".join(
        "%s: $%s" % (field.graphql_name, field.graphql_name)
        for field, _ in values.items()
    )
    params = {
        field.graphql_name: (value, field) for field, value in values.items()
    }
    params[id_param] = (db_object.uid, Entity.uid)

    query_str = """mutation update%sPyApi%s{update%s(
        where: {id: $%s} data: {%s}) {%s}} """ % (
        utils.title_case(type_name),
        format_param_declaration(params),
        type_name,
        id_param,
        values_str,
        results_query_part(type(db_object)),
    )

    return query_str, {name: value for name, (value, _) in params.items()}


def delete(db_object):
    """Generates a query that deletes the given `db_object` from the DB.

    Args:
        db_object (DbObject): The DB object being deleted.
    """
    id_param = "%sId" % db_object.type_name()
    query_str = """mutation delete%sPyApi%s{update%s(
        where: {id: $%s} data: {deleted: true}) {id}} """ % (
        db_object.type_name(),
        "($%s: ID!)" % id_param,
        db_object.type_name(),
        id_param,
    )

    return query_str, {id_param: db_object.uid}


def bulk_delete(db_objects, use_where_clause):
    """Generates a query that bulk-deletes the given `db_objects` from the
    DB.

    Args:
        db_objects (list): A list of DB objects of the same type.
        use_where_clause (bool): If the object IDs should be passed to the
            mutation in a `where` clause or directly as a mutation value.
    """
    type_name = db_objects[0].type_name()
    if use_where_clause:
        query_str = (
            "mutation delete%ssPyApi{delete%ss(where: {%sIds: [%s]}){id}}"
        )
    else:
        query_str = "mutation delete%ssPyApi{delete%ss(%sIds: [%s]){id}}"
    query_str = query_str % (
        utils.title_case(type_name),
        utils.title_case(type_name),
        utils.camel_case(type_name),
        ", ".join('"%s"' % db_object.uid for db_object in db_objects),
    )
    return query_str, {}


def where_as_dict(entity, node: Comparison) -> Dict[str, Any]:
    check_where_clause(entity, node)
    """ Helper that constructs a where clause from a Comparison node. """
    COMPARISON_TO_SUFFIX = {
        Comparison.Op.EQ: "",
        Comparison.Op.NE: "_not",
        Comparison.Op.LT: "_lt",
        Comparison.Op.GT: "_gt",
        Comparison.Op.LE: "_lte",
        Comparison.Op.GE: "_gte",
    }

    key = f"{node.field.graphql_name}{COMPARISON_TO_SUFFIX[node.op]}"
    return {key: node.value}


def order_by_as_string(entity, node: tuple) -> str:
    check_order_by_clause(entity, node)
    return f"{node[0].graphql_name}_{node[1].name.upper()}"



import requests


def test_file_upload(client, rand_gen, dataset):
    data = rand_gen(str)
    uri = client.upload_data(data.encode())
    data_row = dataset.create_data_row(row_data=uri)
    assert requests.get(data_row.row_data).text == data



import uuid
import pytest
from labelbox import parser

from labelbox.schema.annotation_import import AnnotationImportState, LabelImport

"""
- Here we only want to check that the uploads are calling the validation
- Then with unit tests we can check the types of errors raised

"""


def test_create_with_url_arg(
    client, module_project, annotation_import_test_helpers
):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    label_import = LabelImport.create(
        client=client, id=module_project.uid, name=name, url=url
    )
    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name, url)


def test_create_from_url(
    client, module_project, annotation_import_test_helpers
):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    label_import = LabelImport.create_from_url(
        client=client, project_id=module_project.uid, name=name, url=url
    )
    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name, url)


def test_create_with_labels_arg(
    client, module_project, object_predictions, annotation_import_test_helpers
):
    """this test should check running state only to validate running, not completed"""
    name = str(uuid.uuid4())

    label_import = LabelImport.create(
        client=client,
        id=module_project.uid,
        name=name,
        labels=object_predictions,
    )

    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions
    )


def test_create_from_objects(
    client, module_project, object_predictions, annotation_import_test_helpers
):
    """this test should check running state only to validate running, not completed"""
    name = str(uuid.uuid4())

    label_import = LabelImport.create_from_objects(
        client=client,
        project_id=module_project.uid,
        name=name,
        labels=object_predictions,
    )

    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions
    )


def test_create_with_path_arg(
    client,
    tmp_path,
    module_project,
    object_predictions,
    annotation_import_test_helpers,
):
    project = module_project
    name = str(uuid.uuid4())
    file_name = f"{name}.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        parser.dump(object_predictions, f)

    label_import = LabelImport.create(
        client=client, id=project.uid, name=name, path=str(file_path)
    )

    assert label_import.parent_id == project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions
    )


def test_create_from_local_file(
    client,
    tmp_path,
    module_project,
    object_predictions,
    annotation_import_test_helpers,
):
    project = module_project
    name = str(uuid.uuid4())
    file_name = f"{name}.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        parser.dump(object_predictions, f)

    label_import = LabelImport.create_from_file(
        client=client, project_id=project.uid, name=name, path=str(file_path)
    )

    assert label_import.parent_id == project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions
    )


def test_get(client, module_project, annotation_import_test_helpers):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    label_import = LabelImport.create_from_url(
        client=client, project_id=module_project.uid, name=name, url=url
    )

    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name, url)


def test_get_import_jobs_from_project(client, configured_project):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    label_import = LabelImport.create_from_url(
        client=client, project_id=configured_project.uid, name=name, url=url
    )
    label_import.wait_until_done()

    label_imports = list(configured_project.get_label_imports())
    assert len(label_imports) == 1
    assert label_imports[0].input_file_url == url


@pytest.mark.slow
def test_wait_till_done(client, module_project, predictions):
    name = str(uuid.uuid4())
    label_import = LabelImport.create_from_objects(
        client=client,
        project_id=module_project.uid,
        name=name,
        labels=predictions,
    )

    assert len(label_import.inputs) == len(predictions)
    label_import.wait_until_done()

    assert label_import.state == AnnotationImportState.FINISHED
    assert len(label_import.inputs) == len(predictions)
    input_uuids = [input_annot["uuid"] for input_annot in label_import.inputs]
    inference_uuids = [pred["uuid"] for pred in predictions]
    assert set(input_uuids) == set(inference_uuids)
    assert len(label_import.statuses) == len(predictions)
    status_uuids = [
        input_annot["uuid"] for input_annot in label_import.statuses
    ]
    assert set(input_uuids) == set(status_uuids)



from labelbox.schema.tool_building.fact_checking_tool import FactCheckingTool


def test_fact_checking_as_dict_default():
    tool = FactCheckingTool(name="Fact Checking Tool")

    # Get the dictionary representation
    tool_dict = tool.asdict()

    # Expected dictionary structure
    expected_dict = {
        "tool": "fact-checking",
        "name": "Fact Checking Tool",
        "required": False,
        "schemaNodeId": None,
        "featureSchemaId": None,
        "color": None,
        "definition": {
            "variants": [
                {"id": 0, "name": "Accurate", "actions": ["justification"]},
                {"id": 1, "name": "Inaccurate", "actions": ["justification"]},
                {"id": 2, "name": "Disputed", "actions": ["justification"]},
                {
                    "id": 3,
                    "name": "Unsupported",
                    "actions": [],
                },
                {
                    "id": 4,
                    "name": "Can't confidently assess",
                    "actions": [],
                },
                {
                    "id": 5,
                    "name": "No factual information",
                    "actions": [],
                },
            ],
            "version": 1,
        },
    }

    assert tool_dict == expected_dict


def test_step_reasoning_as_dict_with_actions():
    tool = FactCheckingTool(name="Fact Checking Tool")
    for variant in tool.definition.variants:
        variant.set_actions([])

    # Get the dictionary representation
    tool_dict = tool.asdict()

    # Expected dictionary structure
    expected_dict = {
        "tool": "fact-checking",
        "name": "Fact Checking Tool",
        "required": False,
        "schemaNodeId": None,
        "featureSchemaId": None,
        "color": None,
        "definition": {
            "variants": [
                {"id": 0, "name": "Accurate", "actions": []},
                {"id": 1, "name": "Inaccurate", "actions": []},
                {"id": 2, "name": "Disputed", "actions": []},
                {
                    "id": 3,
                    "name": "Unsupported",
                    "actions": [],
                },
                {
                    "id": 4,
                    "name": "Can't confidently assess",
                    "actions": [],
                },
                {
                    "id": 5,
                    "name": "No factual information",
                    "actions": [],
                },
            ],
            "version": 1,
        },
    }

    assert tool_dict == expected_dict



import os
import time
from multiprocessing.dummy import Pool

import lbox.exceptions
import pytest
from google.api_core.exceptions import RetryError

import labelbox.client
from labelbox.schema.media_type import MediaType


def test_missing_api_key():
    key = os.environ.get(lbox.request_client._LABELBOX_API_KEY, None)
    if key is not None:
        del os.environ[lbox.request_client._LABELBOX_API_KEY]

    with pytest.raises(lbox.exceptions.AuthenticationError) as excinfo:
        labelbox.client.Client()

    assert excinfo.value.message == "Labelbox API key not provided"

    if key is not None:
        os.environ[lbox.request_client._LABELBOX_API_KEY] = key


def test_bad_key(rand_gen):
    bad_key = "BAD_KEY_" + rand_gen(str)
    client = labelbox.client.Client(api_key=bad_key)

    with pytest.raises(lbox.exceptions.AuthenticationError) as excinfo:
        client.create_project(name=rand_gen(str), media_type=MediaType.Image)


def test_syntax_error(client):
    with pytest.raises(lbox.exceptions.InvalidQueryError) as excinfo:
        client.execute("asda", check_naming=False)
    assert excinfo.value.message.startswith("Syntax Error:")


def test_semantic_error(client):
    with pytest.raises(lbox.exceptions.InvalidQueryError) as excinfo:
        client.execute("query {bbb {id}}", check_naming=False)
    assert excinfo.value.message.startswith('Cannot query field "bbb"')


def test_timeout_error(client, project):
    with pytest.raises(RetryError) as excinfo:
        query_str = """query getOntology {
        project (where: {id: %s}) {
            ontology {
                normalized
                }
            }
        } """ % (project.uid)

        # Setting connect timeout to 30s, and read timeout to 0.01s
        client.execute(query_str, check_naming=False, timeout=(30.0, 0.01))


def test_query_complexity_error(client):
    with pytest.raises(lbox.exceptions.ValidationFailedError) as excinfo:
        client.execute(
            "{projects {datasets {dataRows {labels {id}}}}}", check_naming=False
        )
    assert excinfo.value.message == "Query complexity limit exceeded"


def test_resource_not_found_error(client):
    with pytest.raises(lbox.exceptions.ResourceNotFoundError):
        client.get_project("invalid project ID")


def test_network_error(client):
    client = labelbox.client.Client(
        api_key=client._request_client.api_key, endpoint="not_a_valid_URL"
    )

    with pytest.raises(lbox.exceptions.NetworkError) as excinfo:
        client.create_project(name="Project name", media_type=MediaType.Image)


@pytest.mark.skip("timeouts cause failure before rate limit")
def test_api_limit_error(client):
    def get(arg):
        try:
            return client.get_user()
        except lbox.exceptions.ApiLimitError as e:
            return e

    # Rate limited at 1500 + buffer
    n = 1600
    # max of 30 concurrency before the service becomes unavailable
    with Pool(30) as pool:
        start = time.time()
        results = list(pool.imap(get, range(n)), total=n)
        elapsed = time.time() - start

    assert elapsed < 60, "Didn't finish fast enough"
    assert lbox.exceptions.ApiLimitError in {type(r) for r in results}

    # Sleep at the end of this test to allow other tests to execute.
    time.sleep(60)



import os

import requests
import pytest
import uuid

from labelbox import Client


@pytest.mark.skip(
    reason="Google credentials are being updated for this test, disabling till it's all sorted out"
)
@pytest.mark.skipif(
    not os.environ.get("DA_GCP_LABELBOX_API_KEY"),
    reason="DA_GCP_LABELBOX_API_KEY not found",
)
def test_default_integration():
    """
    This tests assumes the following:
    1. gcp delegated access is configured to work with jtso-gcs-sdk-da-tests
    2. the integration name is gcs sdk test bucket
    3. This integration is the default

    Currently tests against:
    Org ID: cl269lvvj78b50zau34s4550z
    Email: jtso+gcp_sdk_tests@labelbox.com"""
    client = Client(api_key=os.environ.get("DA_GCP_LABELBOX_API_KEY"))
    ds = client.create_dataset(name="new_ds")
    dr = ds.create_data_row(
        row_data="gs://jtso-gcs-sdk-da-tests/nikita-samokhin-D6QS6iv_CTY-unsplash.jpg"
    )
    assert requests.get(dr.row_data).status_code == 200
    assert ds.iam_integration().name == "gcs sdk test bucket"
    ds.delete()


@pytest.mark.skip(
    reason="Google credentials are being updated for this test, disabling till it's all sorted out"
)
@pytest.mark.skipif(
    not os.environ.get("DA_GCP_LABELBOX_API_KEY"),
    reason="DA_GCP_LABELBOX_API_KEY not found",
)
def test_non_default_integration():
    """
    This tests assumes the following:
    1. aws delegated access is configured to work with lbox-test-bucket
    2. an integration called aws is available to the org

    Currently tests against:
    Org ID: cl26d06tk0gch10901m7jeg9v
    Email: jtso+aws_sdk_tests@labelbox.com
    """
    client = Client(api_key=os.environ.get("DA_GCP_LABELBOX_API_KEY"))
    integrations = client.get_organization().get_iam_integrations()
    integration = [
        inte for inte in integrations if "aws-da-test-bucket" in inte.name
    ][0]
    assert integration.valid
    ds = client.create_dataset(iam_integration=integration, name="new_ds")
    assert ds.iam_integration().name == "aws-da-test-bucket"
    dr = ds.create_data_row(
        row_data="https://jtso-aws-da-sdk-tests.s3.us-east-2.amazonaws.com/adrian-yu-qkN4D3Rf1gw-unsplash.jpg"
    )
    assert requests.get(dr.row_data).status_code == 200
    ds.delete()


def test_no_integration(client, image_url):
    ds = client.create_dataset(iam_integration=None, name="new_ds")
    assert ds.iam_integration() is None
    dr = ds.create_data_row(row_data=image_url)
    assert requests.get(dr.row_data).status_code == 200
    ds.delete()


@pytest.mark.skip(reason="Assumes state of account doesn't have integration")
def test_no_default_integration(client):
    ds = client.create_dataset(name="new_ds")
    assert ds.iam_integration() is None
    ds.delete()


@pytest.mark.skip(
    reason="Google credentials are being updated for this test, disabling till it's all sorted out"
)
@pytest.mark.skipif(
    not os.environ.get("DA_GCP_LABELBOX_API_KEY"),
    reason="DA_GCP_LABELBOX_API_KEY not found",
)
def test_add_integration_from_object():
    """
    This test is based on test_non_default_integration() and assumes the following:

    1. aws delegated access is configured to work with lbox-test-bucket
    2. an integration called aws is available to the org

    Currently tests against:
    Org ID: cl26d06tk0gch10901m7jeg9v
    Email: jtso+aws_sdk_tests@labelbox.com
    """
    client = Client(api_key=os.environ.get("DA_GCP_LABELBOX_API_KEY"))
    integrations = client.get_organization().get_iam_integrations()

    # Prepare dataset with no integration
    integration = [
        integration
        for integration in integrations
        if "aws-da-test-bucket" in integration.name
    ][0]

    ds = client.create_dataset(
        iam_integration=None, name=f"integration_add_obj-{uuid.uuid4()}"
    )

    # Test set integration with object
    new_integration = ds.add_iam_integration(integration)
    assert new_integration == integration

    # Cleaning
    ds.delete()


@pytest.mark.skip(
    reason="Google credentials are being updated for this test, disabling till it's all sorted out"
)
@pytest.mark.skipif(
    not os.environ.get("DA_GCP_LABELBOX_API_KEY"),
    reason="DA_GCP_LABELBOX_API_KEY not found",
)
def test_add_integration_from_uid():
    """
    This test is based on test_non_default_integration() and assumes the following:

    1. aws delegated access is configured to work with lbox-test-bucket
    2. an integration called aws is available to the org

    Currently tests against:
    Org ID: cl26d06tk0gch10901m7jeg9v
    Email: jtso+aws_sdk_tests@labelbox.com
    """
    client = Client(api_key=os.environ.get("DA_GCP_LABELBOX_API_KEY"))
    integrations = client.get_organization().get_iam_integrations()

    # Prepare dataset with no integration
    integration = [
        integration
        for integration in integrations
        if "aws-da-test-bucket" in integration.name
    ][0]

    ds = client.create_dataset(
        iam_integration=None, name=f"integration_add_id-{uuid.uuid4()}"
    )

    # Test set integration with integration id
    integration_id = [
        integration.uid
        for integration in integrations
        if "aws-da-test-bucket" in integration.name
    ][0]

    new_integration = ds.add_iam_integration(integration_id)
    assert new_integration == integration

    # Cleaning
    ds.delete()


@pytest.mark.skip(
    reason="Google credentials are being updated for this test, disabling till it's all sorted out"
)
@pytest.mark.skipif(
    not os.environ.get("DA_GCP_LABELBOX_API_KEY"),
    reason="DA_GCP_LABELBOX_API_KEY not found",
)
def test_integration_remove():
    """
    This test is based on test_non_default_integration() and assumes the following:

    1. aws delegated access is configured to work with lbox-test-bucket
    2. an integration called aws is available to the org

    Currently tests against:
    Org ID: cl26d06tk0gch10901m7jeg9v
    Email: jtso+aws_sdk_tests@labelbox.com
    """
    client = Client(api_key=os.environ.get("DA_GCP_LABELBOX_API_KEY"))
    integrations = client.get_organization().get_iam_integrations()

    # Prepare dataset with an existing integration
    integration = [
        integration
        for integration in integrations
        if "aws-da-test-bucket" in integration.name
    ][0]

    ds = client.create_dataset(
        iam_integration=integration, name=f"integration_remove-{uuid.uuid4()}"
    )

    # Test unset integration
    ds.remove_iam_integration()
    assert ds.iam_integration() is None

    # Cleaning
    ds.delete()



import io
import json
import logging
from typing import Any, Callable, Dict, List, Optional
from urllib.parse import urlparse

import requests
from lbox.exceptions import LabelboxError
from requests import Response, Session

logger = logging.getLogger(__name__)


class AdvClient:
    def __init__(self, endpoint: str, api_key: str):
        self.endpoint = endpoint
        self.api_key = api_key
        self.session = self._create_session()

    def create_embedding(self, name: str, dims: int) -> Dict[str, Any]:
        data = {"name": name, "dims": dims}
        return self._request("POST", "/adv/v1/embeddings", data)

    def delete_embedding(self, id: str):
        return self._request("DELETE", f"/adv/v1/embeddings/{id}")

    def get_embedding(self, id: str) -> Dict[str, Any]:
        return self._request("GET", f"/adv/v1/embeddings/{id}")

    def get_embeddings(self) -> List[Dict[str, Any]]:
        return self._request("GET", "/adv/v1/embeddings").get("results", [])

    def import_vectors_from_file(self, id: str, file_path: str, callback=None):
        self._send_ndjson(
            f"/adv/v1/embeddings/{id}/_import_ndjson", file_path, callback
        )

    def get_imported_vector_count(self, id: str) -> int:
        data = self._request("GET", f"/adv/v1/embeddings/{id}/vectors/_count")
        return data.get("count", 0)

    def _create_session(self) -> Session:
        session = requests.session()
        session.headers.update(
            {
                "Authorization": f"Bearer {self.api_key}",
                "Content-Type": "application/json",
            }
        )
        return session

    def _request(
        self,
        method: str,
        path: str,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        url = f"{self.endpoint}{path}"
        requests_data = None
        if data:
            requests_data = json.dumps(data)
        response = self.session.request(
            method, url, data=requests_data, headers=headers
        )
        if response.status_code != requests.codes.ok:
            message = response.json().get("message")
            if message:
                raise LabelboxError(message)
            else:
                response.raise_for_status()
        return response.json()

    def _send_ndjson(
        self,
        path: str,
        file_path: str,
        callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    ):
        """
        Sends an NDJson file in chunks.

        Args:
            path: The URL path
            file_path: The path to the NDJSON file.
            callback: A callback to run for each chunk uploaded.
        """

        def upload_chunk(_buffer, _count):
            _buffer.write(b"\n")
            _headers = {
                "Content-Type": "application/x-ndjson",
                "X-Content-Lines": str(_count),
                "Content-Length": str(buffer.tell()),
            }
            rsp = self._send_bytes(f"{self.endpoint}{path}", _buffer, _headers)
            rsp.raise_for_status()
            if callback:
                callback(rsp.json())

        buffer = io.BytesIO()
        count = 0
        with open(file_path, "rb") as fp:
            for line in fp:
                buffer.write(line)
                count += 1
                if count >= 1000:
                    upload_chunk(buffer, count)
                    buffer = io.BytesIO()
                    count = 0
        if count:
            upload_chunk(buffer, count)

    def _send_bytes(
        self,
        url: str,
        buffer: io.BytesIO,
        headers: Optional[Dict[str, Any]] = None,
    ) -> Response:
        buffer.seek(0)
        return self.session.put(url, headers=headers, data=buffer)

    @classmethod
    def factory(cls, api_endpoint: str, api_key: str) -> "AdvClient":
        parsed_url = urlparse(api_endpoint)
        endpoint = f"{parsed_url.scheme}://{parsed_url.netloc}/adv"
        return AdvClient(endpoint, api_key)



import json

from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.mixins import CustomMetric

from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.types import Label, ObjectAnnotation, TextEntity


def test_text_entity_import():
    with open("tests/data/assets/ndjson/text_entity_import.json", "r") as file:
        data = json.load(file)

    labels = [
        Label(
            data=GenericDataRowData(
                uid="cl6xnv9h61fv0085yhtoq06ht",
            ),
            annotations=[
                ObjectAnnotation(
                    custom_metrics=[
                        CustomMetric(name="customMetric1", value=0.5),
                        CustomMetric(name="customMetric2", value=0.3),
                    ],
                    confidence=0.53,
                    name="some-text-entity",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "5ad9c52f-058d-49c8-a749-3f20b84f8cd4",
                    },
                    value=TextEntity(start=67, end=128, extra={}),
                )
            ],
        )
    ]
    res = list(NDJsonConverter.serialize(labels))
    assert res == data


def test_text_entity_import_without_confidence():
    with open(
        "tests/data/assets/ndjson/text_entity_without_confidence_import.json",
        "r",
    ) as file:
        data = json.load(file)

    labels = [
        Label(
            data=GenericDataRowData(
                uid="cl6xnv9h61fv0085yhtoq06ht",
            ),
            annotations=[
                ObjectAnnotation(
                    name="some-text-entity",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "5ad9c52f-058d-49c8-a749-3f20b84f8cd4",
                    },
                    value=TextEntity(start=67, end=128, extra={}),
                )
            ],
        )
    ]

    res = list(NDJsonConverter.serialize(labels))
    assert res == data



from typing import Dict, Any, Optional, Union, Tuple
from abc import ABC, abstractmethod

import geojson
import numpy as np

from shapely import geometry as geom
from pydantic import BaseModel


class Geometry(BaseModel, ABC):
    """Abstract base class for geometry objects"""

    extra: Dict[str, Any] = {}

    @property
    def shapely(
        self,
    ) -> Union[
        geom.Point,
        geom.LineString,
        geom.Polygon,
        geom.MultiPoint,
        geom.MultiLineString,
        geom.MultiPolygon,
    ]:
        return geom.shape(self.geometry)

    def get_or_create_canvas(
        self,
        height: Optional[int],
        width: Optional[int],
        canvas: Optional[np.ndarray],
    ) -> np.ndarray:
        if canvas is None:
            if height is None or width is None:
                raise ValueError(
                    "Must either provide canvas or height and width"
                )
            canvas = np.zeros((height, width, 3), dtype=np.uint8)
        canvas = np.ascontiguousarray(canvas)
        return canvas

    @property
    @abstractmethod
    def geometry(self) -> geojson:
        pass

    @abstractmethod
    def draw(
        self,
        height: Optional[int] = None,
        width: Optional[int] = None,
        canvas: Optional[np.ndarray] = None,
        color: Optional[Union[int, Tuple[int, int, int]]] = None,
        thickness: Optional[int] = 1,
    ) -> np.ndarray:
        pass



import uuid
import pytest
from unittest.mock import MagicMock, patch

from labelbox.schema.annotation_import import MALPredictionImport, logger


def test_should_warn_user_about_unsupported_confidence():
    """this test should check running state only to validate running, not completed"""
    id = str(uuid.uuid4())

    labels = [
        {
            "bbox": {"height": 428, "left": 2089, "top": 1251, "width": 158},
            "classifications": [
                {
                    "answer": [
                        {
                            "schemaId": "ckrb1sfl8099e0y919v260awv",
                            "confidence": 0.894,
                        }
                    ],
                    "schemaId": "ckrb1sfkn099c0y910wbo0p1a",
                }
            ],
            "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
            "schemaId": "ckrb1sfjx099a0y914hl319ie",
            "uuid": "d009925d-91a3-4f67-abd9-753453f5a584",
        },
    ]
    with patch.object(MALPredictionImport, "_create_mal_import_from_bytes"):
        with patch.object(logger, "warning") as warning_mock:
            MALPredictionImport.create_from_objects(
                client=MagicMock(), project_id=id, name=id, predictions=labels
            )
            warning_mock.assert_called_once()
            "Confidence scores are not supported in MAL Prediction Import" in warning_mock.call_args_list[
                0
            ].args[0]


def test_invalid_labels_format():
    """this test should confirm that annotations are required to be in a form of list"""
    id = str(uuid.uuid4())

    label = {
        "bbox": {"height": 428, "left": 2089, "top": 1251, "width": 158},
        "classifications": [
            {
                "answer": [
                    {
                        "schemaId": "ckrb1sfl8099e0y919v260awv",
                        "confidence": 0.894,
                    }
                ],
                "schemaId": "ckrb1sfkn099c0y910wbo0p1a",
            }
        ],
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "schemaId": "ckrb1sfjx099a0y914hl319ie",
        "uuid": "3a83db52-75e0-49af-a171-234ce604502a",
    }

    with patch.object(MALPredictionImport, "_create_mal_import_from_bytes"):
        with pytest.raises(TypeError):
            MALPredictionImport.create_from_objects(
                client=MagicMock(), project_id=id, name=id, predictions=label
            )



import pytest
from lbox.exceptions import InvalidQueryError

from labelbox import Project
from labelbox.schema.media_type import MediaType


@pytest.fixture
def project_to_test_where(client, rand_gen):
    p_a_name = f"a-{rand_gen(str)}"
    p_b_name = f"b-{rand_gen(str)}"
    p_c_name = f"c-{rand_gen(str)}"

    p_a = client.create_project(name=p_a_name, media_type=MediaType.Image)
    p_b = client.create_project(name=p_b_name, media_type=MediaType.Image)
    p_c = client.create_project(name=p_c_name, media_type=MediaType.Image)

    yield p_a, p_b, p_c

    p_a.delete()
    p_b.delete()
    p_c.delete()


# Avoid assertions using equality to prevent intermittent failures due to
# other builds simultaneously adding projects to test org
@pytest.mark.skip(reason="Flake, rarely used functionality")
def test_where(client, project_to_test_where):
    p_a, p_b, p_c = project_to_test_where
    p_a_name = p_a.name
    p_b_name = p_b.name

    def get(where=None):
        date_where = Project.created_at >= min(
            p_a.created_at, p_b.created_at, p_c.created_at
        )
        where = date_where if where is None else where & date_where
        return {p.uid for p in client.get_projects(where)}

    assert {p_a.uid, p_b.uid, p_c.uid}.issubset(get())
    e_a = get(Project.name == p_a_name)
    assert p_a.uid in e_a and p_b.uid not in e_a and p_c.uid not in e_a
    not_b = get(Project.name != p_b_name)
    assert {p_a.uid, p_c.uid}.issubset(not_b) and p_b.uid not in not_b
    gt_b = get(Project.name > p_b_name)
    assert p_c.uid in gt_b and p_a.uid not in gt_b and p_b.uid not in gt_b
    lt_b = get(Project.name < p_b_name)
    assert p_a.uid in lt_b and p_b.uid not in lt_b and p_c.uid not in lt_b
    ge_b = get(Project.name >= p_b_name)
    assert {p_b.uid, p_c.uid}.issubset(ge_b) and p_a.uid not in ge_b


def test_unsupported_where(client):
    with pytest.raises(InvalidQueryError):
        client.get_projects(where=(Project.name == "a") & (Project.name == "b"))

    # TODO support logical OR and NOT in where
    with pytest.raises(InvalidQueryError):
        client.get_projects(
            where=(Project.name == "a") | (Project.description == "b")
        )

    with pytest.raises(InvalidQueryError):
        client.get_projects(where=~(Project.name == "a"))



from labelbox import Project, Dataset
from labelbox.orm import query
from labelbox.orm.comparison import Comparison, LogicalExpression


def format(*args, **kwargs):
    return query.Query(*args, **kwargs).format()[0]


def test_query_what():
    assert format("first", Project).startswith("first{")
    assert format("other", Project).startswith("other{")


def test_query_subquery():
    assert format("x", query.Query("sub", Project)).startswith("x{sub{")
    assert format("x", query.Query("bus", Project)).startswith("x{bus{")


def test_query_where():
    q, p = query.Query("x", Project, Project.name > "name").format()
    assert q.startswith("x(where: {name_gt: $param_0}){")
    assert p == {"param_0": ("name", Project.name)}

    q, p = query.Query(
        "x", Project, (Project.name != "name") & (Project.uid <= 42)
    ).format()
    assert q.startswith(
        "x(where: {AND: [{name_not: $param_0}, {id_lte: $param_1}]}"
    )
    assert p == {
        "param_0": ("name", Project.name),
        "param_1": (42, Project.uid),
    }


def test_query_param_declaration():
    q, _ = query.Query("x", Project, Project.name > "name").format_top("y")
    assert q.startswith("query yPyApi($param_0: String!){x")

    q, _ = query.Query(
        "x", Project, (Project.name > "name") & (Project.uid == 42)
    ).format_top("y")
    assert q.startswith("query yPyApi($param_0: String!, $param_1: ID!){x")


def test_query_order_by():
    q, _ = query.Query("x", Project, order_by=Project.name.asc).format()
    assert q.startswith("x(orderBy: name_ASC){")

    q, _ = query.Query("x", Project, order_by=Project.uid.desc).format()
    assert q.startswith("x(orderBy: id_DESC){")


def test_logical_ops():
    Op = LogicalExpression.Op

    assert list(query.logical_ops(None)) == []
    comparison_1 = Comparison.Op.EQ(Project.name, "name")
    assert list(query.logical_ops(comparison_1)) == []
    comparison_2 = Comparison.Op.LT(Dataset.uid, "uid")
    assert list(query.logical_ops(comparison_2)) == []
    op_1 = Op.AND(comparison_1, comparison_2)
    assert list(query.logical_ops(op_1)) == [Op.AND]
    op_2 = Op.OR(comparison_1, comparison_2)
    assert list(query.logical_ops(op_2)) == [Op.OR]
    op_3 = Op.OR(op_1, op_2)
    assert list(query.logical_ops(op_3)) == [Op.OR, Op.AND, Op.OR]
